"""
### Perform convolutions on arrays.  ###
"""

# Standard Library
from typing import Tuple

# External
import numpy as np
from numba import jit



@jit(nopython=True, nogil=True, cache=True, fastmath=True, inline='always')
def _simple_blur_kernel_2d_3x3() -> Tuple[np.ndarray, np.ndarray]:
    """
    Create a 2D blur kernel.

    The kernel has the following form:
    ```python
    >>> weights = [
    ...     0.08422299, 0.12822174, 0.08422299,
    ...     0.12822174, 0.15022110, 0.12822174,
    ...     0.08422299, 0.12822174, 0.08422299,
    ... ]
    ```
    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        A tuple of two numpy arrays. The first array contains the (x, y) offsets
        of the kernel values. The second array contains the corresponding weights
        of each kernel value.
    """
    offsets = np.array([
        [ 1, -1], [ 1, 0], [ 1, 1],
        [ 0, -1], [ 0, 0], [ 0, 1],
        [-1, -1], [-1, 0], [-1, 1],
    ])

    weights = np.array([
        0.08422299, 0.12822174, 0.08422299,
        0.12822174, 0.1502211 , 0.12822174,
        0.08422299, 0.12822174, 0.08422299,
    ], dtype=np.float32)

    return offsets, weights


@jit(nopython=True, nogil=True, cache=True, fastmath=True, inline='always')
def _simple_unsharp_kernel_2d_3x3() -> Tuple[np.ndarray, np.ndarray]:
    """
    Create a 2D unsharp kernel.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        A tuple of two numpy arrays. The first array contains the (x, y) offsets
        of the kernel values. The second array contains the corresponding weights
        of each kernel value.
    """
    offsets = np.array([
        [ 1, -1], [ 1, 0], [ 1, 1],
        [ 0, -1], [ 0, 0], [ 0, 1],
        [-1, -1], [-1, 0], [-1, 1],
    ])

    weights = np.array([
        -0.09911165, -0.15088834, -0.09911165,
        -0.15088834,  2.        , -0.15088834,
        -0.09911165, -0.15088834, -0.09911165,
    ], dtype=np.float32)

    return offsets, weights


@jit(nopython=True, nogil=True, cache=True, fastmath=True, inline='always')
def _simple_shift_kernel_2d(
    x_offset: float,
    y_offset: float,
) -> Tuple[np.ndarray, np.ndarray]:
    """ Create a 2D shift kernel. For augmentations. """
    return kernel_shift(x_offset, y_offset)


@jit(nopython=True, nogil=True, cache=True, fastmath=True, inline="always")
def _distance_2D(p1: np.ndarray, p2: np.ndarray) -> float:
    """ Returns the distance between two points. (2D) """
    d1 = (p1[0] - p2[0]) ** 2
    d2 = (p1[1] - p2[1]) ** 2

    return np.sqrt(d1 + d2)


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def _area_covered(square, radius):
    """
    Calculates the area covered by a circle within a square.
    Monte-carlo(ish) method. Can be parallelized.
    """
    n_points = 100
    min_y = square[:, 0].min()
    max_y = square[:, 0].max()
    min_x = square[:, 1].min()
    max_x = square[:, 1].max()

    steps = int(np.rint(np.sqrt(n_points)))
    range_y = np.linspace(min_x, max_x, steps)
    range_x = np.linspace(min_y, max_y, steps)

    center = np.array([0.0, 0.0], dtype=np.float32)
    adjusted_radius = radius + 0.5

    points_within = 0
    for y in range_y:
        for x in range_x:
            point = np.array([y, x], dtype=np.float32)
            if _distance_2D(center, point) <= adjusted_radius:
                points_within += 1

    area = points_within / (steps ** 2)

    return area


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def _circular_kernel_2D(radius):
    """ Creates a circular 2D kernel. Supports fractional radii. """
    size = np.int64(np.ceil(radius) * 2 + 1)
    kernel = np.zeros((size, size), dtype=np.float32)

    center = np.array([0.0, 0.0], dtype=np.float32)

    step = size // 2
    for idx_i, col in enumerate(range(-step, step + 1)):
        for idx_j, row in enumerate(range(-step, step + 1)):
            square = np.zeros((4, 2), dtype=np.float32)
            square[0] = np.array([col - 0.5, row - 0.5], dtype=np.float32)
            square[1] = np.array([col + 0.5, row - 0.5], dtype=np.float32)
            square[2] = np.array([col + 0.5, row + 0.5], dtype=np.float32)
            square[3] = np.array([col - 0.5, row + 0.5], dtype=np.float32)

            within = np.zeros(4, dtype=np.uint8)
            for i in range(4):
                within[i] = _distance_2D(center, square[i]) <= radius + 0.5

            # Case 1: completely inside
            if within.sum() == 4:
                kernel[idx_i][idx_j] = 1.0

            # Case 2: completely outside
            elif within.sum() == 0:
                kernel[idx_i][idx_j] = 0.0

            # Case 3: partially inside
            else:
                kernel[idx_i][idx_j] = _area_covered(square, radius)

    return kernel


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def _distance_weighted_kernel_2D(radius, method, decay=0.2, sigma=2.0):
    """
    Creates a distance weighted kernel.
    
    Parameters
    ----------

    radius : float
        Radius of the kernel.
    
    method : int
        Method to use for weighting.
        0. linear
        1. sqrt
        2. power
        3. log
        4. gaussian
        5. constant
    """
    size = np.int64(np.ceil(radius) * 2 + 1)
    kernel = np.zeros((size, size), dtype=np.float32)

    center = np.array([0.0, 0.0], dtype=np.float32)

    step = size // 2
    for idx_i, col in enumerate(range(-step, step + 1)):
        for idx_j, row in enumerate(range(-step, step + 1)):
            point = np.array([col, row], dtype=np.float32)
            distance = _distance_2D(center, point)

            # Linear
            if method == 0:
                kernel[idx_i, idx_j] = np.power((1 - decay), distance)

            # Sqrt
            elif method == 1:
                kernel[idx_i, idx_j] = np.power(np.sqrt((1 - decay)), distance)

            # Power
            elif method == 2:
                kernel[idx_i, idx_j] = np.power(np.power((1 - decay), 2), distance)

            # Gaussian
            elif method == 3:
                kernel[idx_i, idx_j] = np.exp(-(np.power(distance, 2)) / (2 * np.power(sigma, 2)))

            # Constant
            else:
                kernel[idx_i, idx_j] = 1.0

    return kernel


def kernel_base(
    radius: float,
    circular: bool = False,
    distance_weighted: bool = False,
    normalised: bool = True,
    hole: bool = False,
    method: int = 0,
    decay: float = 0.2,
    sigma: float = 2.0,
) -> np.ndarray:
    """
    Creates a 2D kernel for convolution.

    Parameters
    ----------
    radius : float
        Radius of the kernel.
    
    circular : bool
        Whether to use a circular kernel.
    
    distance_weighted : bool
        Whether to use a distance weighted kernel.
    
    normalised : bool
        Whether to normalise the kernel.
    
    hole : bool
        Whether to create a hole in the center of the kernel.

    method : int
        Method to use for weighting.
        0. linear
        1. sqrt
        2. power
        3. gaussian
        4. constant
    
    decay : float
        Decay rate for distance weighted kernels. Only used if `distance_weighted` is True.

    sigma : float
        Sigma for gaussian distance weighted kernels. Only used if `distance_weighted` is True and `method` is 3.

    Returns
    -------
    kernel : np.ndarray
        The kernel.
    """
    size = np.int64(np.ceil(radius) * 2 + 1)
    kernel = np.ones((size, size), dtype=np.float32)

    if hole:
        kernel[size // 2, size // 2] = 0.0

    if circular:
        circular_kernel = _circular_kernel_2D(radius)
        kernel *= circular_kernel

    if distance_weighted:
        distance_weighted_kernel = _distance_weighted_kernel_2D(
            radius,
            method,
            decay,
            sigma,
        )
        kernel *= distance_weighted_kernel

    if normalised:
        kernel /= np.sum(kernel)

    return kernel



@jit(nopython=True, nogil=True, cache=True, fastmath=True, inline='always')
def kernel_shift(
    x_offset: float,
    y_offset: float,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Create a 2D shift kernel.

    This function returns a kernel that can be used to shift a raster by a fractional
    number of pixels in the x and y directions. The kernel can also be used to simulate
    channel misalignment in image augmentation.

    Parameters
    ----------
    x_offset : float
        The horizontal (x) offset to apply.

    y_offset : float
        The vertical (y) offset to apply.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        A tuple of two numpy arrays. The first array contains the (x, y) offsets
        of the kernel values. The second array contains the corresponding weights
        of each kernel value.
    """
    if x_offset == 0.0 and y_offset == 0.0:
        offsets = np.array([[0, 0]], dtype=np.int64)
        weights = np.array([1.0], dtype=np.float32)

        return offsets, weights

    y0 = [int(np.floor(y_offset)), int(np.ceil(y_offset))] if y_offset != 0 else [0, 0]
    x0 = [int(np.floor(x_offset)), int(np.ceil(x_offset))] if x_offset != 0 else [0, 0]

    if x_offset == 0.0 or x_offset % 1 == 0.0:
        offsets = np.zeros((2, 2), dtype=np.int64)
        weights = np.zeros(2, dtype=np.float32)

        offsets[0] = [int(x_offset) if x_offset % 1 == 0.0 else 0, y0[0]]
        offsets[1] = [int(x_offset) if x_offset % 1 == 0.0 else 0, y0[1]]

        weights[0] = y_offset - y0[0]
        weights[1] = 1 - weights[0]

    elif y_offset == 0.0 or y_offset % 1 == 0.0:
        offsets = np.zeros((2, 2), dtype=np.int64)
        weights = np.zeros(2, dtype=np.float32)

        offsets[0] = [x0[0], int(y_offset) if y_offset % 1 == 0.0 else 0]
        offsets[1] = [x0[1], int(y_offset) if y_offset % 1 == 0.0 else 0]

        weights[0] = x_offset - x0[0]
        weights[1] = 1 - weights[0]

    else:
        offsets = np.zeros((4, 2), dtype=np.int64)
        weights = np.zeros(4, dtype=np.float32)

        offsets[0] = [x0[0], y0[0]]
        offsets[1] = [x0[0], y0[1]]
        offsets[2] = [x0[1], y0[0]]
        offsets[3] = [x0[1], y0[1]]

        weights[0] = (1 - (x_offset - offsets[0][0])) * (1 - (y_offset - offsets[0][1]))
        weights[1] = (1 - (x_offset - offsets[1][0])) * (1 + (y_offset - offsets[1][1]))
        weights[2] = (1 + (x_offset - offsets[2][0])) * (1 - (y_offset - offsets[2][1]))
        weights[3] = (1 + (x_offset - offsets[3][0])) * (1 + (y_offset - offsets[3][1]))

    return offsets, weights


@jit(nopython=True, nogil=True, cache=True, fastmath=True)
def kernel_unsharp(
    radius: float = 1.0,
    intensity: float = 1.0,
) -> np.ndarray:
    """
    Create a 2D unsharp kernel.

    This function returns a kernel that can be used to apply an unsharp mask to an image.

    Parameters
    ----------
    radius : float, optional
        The radius of the kernel. Default: 1.0.

    intensity : float, optional
        The intensity of the unsharp mask. Default: 1.0.

    Returns
    -------
    np.ndarray
        The kernel.
    """
    kernel = kernel_base(
        radius=radius,
        circular=True,
        distance_weighted=True,
        method=3,
        normalised=True,
        hole=True,
    )

    kernel *= intensity
    kernel_sum = np.sum(kernel) + 1.0
    kernel *= -1.0
    kernel[kernel.shape[0] // 2, kernel.shape[1] // 2] = kernel_sum

    return kernel


@jit(nopython=True, nogil=True, cache=True, fastmath=True, parallel=True)
def kernel_sobel(
    radius=1,
    scale=2,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Creates a 2D Sobel style kernel consisting of a horizontal and vertical component.
    This function returns a kernel that can be used to apply a Sobel filter to an image.

    `kernel_gx, kernel_gy = get_kernel_sobel(radius=1, scale=2)`

    The kernels for radis=2, scale=2 are:
    ```python
    gx = [
        [ 0.56  0.85  0.   -0.85 -0.56],
        [ 0.85  1.5   0.   -1.5  -0.85],
        [ 1.    2.    0.   -2.   -1.  ],
        [ 0.85  1.5   0.   -1.5  -0.85],
        [ 0.56  0.85  0.   -0.85 -0.56],
    ]

    gy = [
        [ 0.56  0.85  1.    0.85  0.56],
        [ 0.85  1.5   2.    1.5   0.85],
        [ 0.    0.    0.    0.    0.  ],
        [-0.85 -1.5  -2.   -1.5  -0.85],
        [-0.56 -0.85 -1.   -0.85 -0.56],
    ]
    ```

    Parameters
    ----------
    radius : float, optional
        The radius of the kernel. Default: 1.0.

    scale : float, optional
        The scale of the kernel. Default: 2.0.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        The kernels 
    
    """
    size = np.int64(np.ceil(radius) * 2 + 1)
    kernel = np.zeros((size, size), dtype=np.float32)
    center = np.array([0.0, 0.0], dtype=np.float32)

    step = size // 2
    for idx_i, col in enumerate(range(-step, step + 1)):
        for idx_j, row in enumerate(range(-step, step + 1)):
            point = np.array([col, row], dtype=np.float32)
            distance = _distance_2D(center, point)
            if col == 0 and row == 0:
                kernel[idx_i, idx_j] = 0
            else:
                weight = np.power((1 - 0.5), distance) * 2
                kernel[idx_i, idx_j] = weight * scale

    # vertical
    kernel_gx = kernel.copy()
    kernel_gx[:, size // 2:] *= -1
    kernel_gx[:, size // 2] = 0

    # horisontal
    kernel_gy = kernel.copy()
    kernel_gy[size // 2:, :] *= -1
    kernel_gy[size // 2, :] = 0

    return kernel_gx, kernel_gy


@jit(nopython=True, nogil=True, cache=True, fastmath=True)
def kernel_get_offsets_and_weights(
    kernel: np.ndarray,
    remove_zero_weights: bool = True,
) -> Tuple[np.ndarray, np.ndarray, int]:
    """
    Generates a list of offsets and weights for a given kernel.

    Parameters
    ----------
    kernel : np.ndarray
        The kernel to generate offsets and weights for.

    remove_zero_weights : bool, optional
        Whether to remove offsets and weights with zero weights. Default: True.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        The offsets and weights
    """
    count = np.count_nonzero(kernel) if remove_zero_weights else kernel.size

    offsets = np.zeros((count, 2), dtype=np.int64)
    weights = np.zeros((count), dtype=np.float32)

    step = kernel.shape[0] // 2
    index = 0
    for col in range(-step, step + 1):
        for row in range(-step, step + 1):
            if kernel[col + step, row + step] != 0.0:
                offsets[index][0] = col
                offsets[index][1] = row
                weights[index] = kernel[col + step, row + step]

                index += 1

    return offsets, weights



"""
### Basic IO functions for working with Rasters ###
"""
# Standard library
import sys; sys.path.append("../../")
from typing import List, Optional, Union, Tuple, Type
import warnings

# External
import numpy as np
from osgeo import gdal, ogr, osr

# Internal
from buteo.utils import (
    utils_base,
    utils_gdal,
    utils_bbox,
    utils_path,
    utils_io,
    utils_translate,
    utils_projection,
)
from buteo.raster import core_raster, core_offsets



def raster_to_array(
    raster: Union[gdal.Dataset, str, List[Union[str, gdal.Dataset]]],
    *,
    bands: Union[List[int], str, int] = 'all',
    filled: bool = False,
    fill_value: Optional[Union[int, float]] = None,
    pixel_offsets: Optional[Union[List[int], Tuple[int, int, int, int]]] = None,
    bbox: Optional[List[float]] = None,
    bbox_srs: Optional[Union[str, osr.SpatialReference]] = None,
    cast: Optional[Union[np.dtype, str]] = None,
    channel_last: bool = True,
) -> np.ndarray:
    """
    Converts a raster or a list of rasters into a NumPy array.

    Parameters
    ----------
    raster : gdal.Dataset or str or list
        Raster(s) to convert. The rasters must be aligned.

    bands : list or str or int, optional
        Bands from the raster to convert to a numpy array. Can be "all", an int,
        or a list of integers, or a single integer. Please note that bands are 1-indexed.
        Default: "all".

    filled : bool, optional
        If True, nodata values in the array will be filled with the specified fill_value.
        If False, a masked array will be created with nodata values masked. Default: False.

    fill_value : int or float, optional
        Value to fill the array with if filled is True. If None, the nodata value
        of the raster is used. Default: None.

    bbox : list, optional
        A list of `[xmin, xmax, ymin, ymax]` to use as the extent of the raster.
        Uses coordinates and the OGR format. Default: None.

    bbox_srs : str or osr.SpatialReference, optional
        The spatial reference of the bounding box. If None, the spatial reference
        of the raster is used. Default: None.

    pixel_offsets : list or tuple, optional
        A list of `[x_offset, y_offset, x_size, y_size]` to use as the extent of the
        raster. Uses pixel offsets and the OGR format. Default: None.

    cast : str or dtype, optional
        A type to cast the array to. If None, the array is not cast. It is only cast
        if the array is not already the dtype. Default: None.

    channel_last : bool, optional
        If True, the output array will have shape (height, width, channels).
        If False, the output array will have shape (channels, height, width).
        Default: True.

    Returns
    -------
    np.ndarray
        A numpy array representing the raster data.
    
    Raises
    ------
    ValueError
        If the raster is not a valid raster.
    ValueError
        If the bands are not valid.
    ValueError
        If the masked parameter is not valid.
    ValueError
        If both bbox and pixel_offsets are provided.
    
    Examples
    --------
    `Example 1: Convert a raster to a numpy array.`
    ```python
    >>> import buteo as beo
    >>> 
    >>> raster = "/path/to/raster/raster.tif"
    >>> 
    >>> # Convert a raster to a numpy array
    >>> array = beo.raster_to_array(raster)
    >>> 
    >>> array.shape, array.dtype
    >>> (100, 100, 3), dtype('uint8')
    ```
    `Example 2: Convert a raster to a numpy array with a specific band.`
    ```python
    >>> import buteo as beo
    >>> 
    >>> raster = "/path/to/raster/raster.tif"
    >>> 
    >>> # Convert a raster to a numpy array
    >>> array = beo.raster_to_array(raster, bands=[2])
    >>> 
    >>> array.shape, array.dtype
    >>> (100, 100, 1), dtype('uint8')
    ```
    `Example 3: Convert a list of rasters to a numpy array with a specific`
    ```python
    >>> # band and a specific type and filled.
    >>> from glob import glob
    >>> import buteo as beo
    >>> 
    >>> FOLDER = "/path/to/folder"
    >>> rasters = glob(FOLDER + "/*.tif")
    >>>
    >>> len(rasters)
    >>> 10
    >>> 
    >>> # Convert rasters to a numpy array
    >>> array = beo.raster_to_array(
    ...     rasters,
    ...     bands=[2],
    ...     cast="float32",
    ...     filled=True,
    ...     fill_value=0.0,
    ... )
    >>> 
    >>> # This raises an error if the 10 rasters are not aligned.
    >>> array.shape, array.dtype
    >>> (100, 100, 10), dtype('float32')
    ```
    """
    utils_base._type_check(raster, [str, gdal.Dataset, [str, gdal.Dataset]], "raster")
    utils_base._type_check(bands, [int, [int], str], "bands")
    utils_base._type_check(filled, [bool], "filled")
    utils_base._type_check(fill_value, [int, float, None], "fill_value")
    utils_base._type_check(bbox, [list, None], "bbox")
    utils_base._type_check(bbox_srs, [str, osr.SpatialReference, None], "bbox_srs")
    utils_base._type_check(pixel_offsets, [list, tuple, None], "pixel_offsets")
    utils_base._type_check(cast, [np.dtype, str, type(np.int64), None], "cast")
    utils_base._type_check(channel_last, [bool], "channel_last")

    input_is_list = isinstance(raster, list)

    raster = utils_io._get_input_paths(raster, "raster")

    assert core_raster.check_rasters_are_aligned(raster), "Rasters are not aligned."

    # Read metadata
    metadata = core_raster._get_basic_metadata_raster(raster[0])
    shape = metadata["shape"]
    dtype = metadata["dtype"] if cast is None else cast
    dtype = utils_translate._parse_dtype(dtype)
    has_nodata = metadata["nodata"]
    nodata_value = np.array(metadata["nodata_value"], dtype=dtype) if has_nodata else None

    # Determine output shape
    x_offset, y_offset, x_size, y_size = 0, 0, shape[1], shape[0]

    if pixel_offsets is not None and bbox is not None:
        raise ValueError("Cannot provide both pixel offsets and a bounding box.")

    elif pixel_offsets is not None:
        x_offset, y_offset, x_size, y_size = pixel_offsets

        if x_offset < 0 or y_offset < 0:
            raise ValueError("Pixel offsets cannot be negative.")

        if x_offset + x_size > shape[1] or y_offset + y_size > shape[0]:
            raise ValueError("Pixel offsets are outside of raster.")

    elif bbox is not None:
        if bbox_srs is not None:
            bbox = utils_projection.reproject_bbox(bbox, bbox_srs, metadata["projection_wkt"])

        if not utils_bbox._check_bboxes_intersect(metadata["bbox"], bbox):
            raise ValueError("Extent is outside of raster.")

        x_offset, y_offset, x_size, y_size = utils_bbox._get_pixel_offsets(metadata["geotransform"], bbox)

        if x_offset < 0 or y_offset < 0:
            raise ValueError("Pixel offsets cannot be negative.")

    if len(raster) > 1 and bands not in ["all", -1]:
        raise ValueError("Cannot specify bands for multiple rasters.")

    bands_to_process = []
    output_shape = shape
    if (isinstance(bands, str) and bands.lower() == "all") or bands == -1:
        total_channels = core_raster._raster_count_bands_list(raster)
        output_shape = [y_size, x_size, total_channels]

        for r in raster:
            meta = core_raster._get_basic_metadata_raster(r)
            bands_in_raster = meta["bands"]
            bands_to_process.append(
                utils_gdal._convert_to_band_list(-1, bands_in_raster),
            )
    else:
        channels = 0
        for r in raster:
            meta = core_raster._get_basic_metadata_raster(r)
            bands_in_raster = meta["bands"]
            bands_in_raster_list = utils_gdal._convert_to_band_list(bands, bands_in_raster)
            bands_to_process.append(bands_in_raster_list)

            channels += len(bands_in_raster_list)

        output_shape = [y_size, x_size, channels]

    # Read data
    output_array = np.ma.zeros(output_shape, dtype=dtype)
    channel = 0
    for idx, r_path in enumerate(raster):

        # We can read all at once
        if len(raster) == 1 and (bands == "all" or bands == -1) and False == True:
            r_open = core_raster._raster_open(r_path)
            data = r_open.ReadAsArray(x_offset, y_offset, x_size, y_size)

            if data.ndim != 3:
                data = np.expand_dims(data, axis=0)

            data = np.transpose(data, (1, 2, 0))

            if cast is not None:
                data = utils_translate._safe_numpy_casting(data, dtype)

            output_array = data
        
        # We need to read bands one by one
        else:
            for n_band in bands_to_process[idx]:
                r_open = core_raster._raster_open(r_path)

                band = r_open.GetRasterBand(n_band)
                data = band.ReadAsArray(x_offset, y_offset, x_size, y_size)

                if cast is not None:
                    output_array[:, :, channel] = utils_translate._safe_numpy_casting(data, dtype)
                else:
                    output_array[:, :, channel] = data

                channel += 1

    if has_nodata:
        output_array = np.ma.masked_equal(output_array, nodata_value)

        if filled:
            if fill_value is None:
                fill_value = nodata_value

                if not utils_translate._check_is_value_within_dtype_range(fill_value, dtype):
                    warnings.warn(
                        f"Fill value {fill_value} is outside of dtype {dtype} range. "
                        "Setting fill value to 0."
                    )
                    fill_value = 0

            output_array = np.ma.getdata(output_array.filled(fill_value))

    elif filled:
        np.nan_to_num(data, nan=fill_value, copy=False)

    else:
        output_array = np.ma.getdata(output_array)

    if filled and np.ma.isMaskedArray(output_array):
        output_array = np.ma.getdata(output_array.filled(fill_value))

    # Reshape array
    if not channel_last:
        output_array = np.transpose(output_array, (2, 0, 1))

    if input_is_list:
        return output_array

    return output_array


def array_to_raster(
    array: np.ndarray,
    *,
    reference: Union[str, gdal.Dataset],
    out_path: Optional[str] = None,
    set_nodata: Optional[Union[float, int, str, bool]] = "arr",
    allow_mismatches: bool = False,
    pixel_offsets: Optional[Union[List[int], Tuple[int, int, int, int]]] = None,
    bbox: Optional[List[float]] = None,
    bbox_srs: Optional[Union[str, osr.SpatialReference]] = None,
    channel_last: bool = True,
    overwrite: bool = True,
    creation_options: Optional[List[str]] = None,
):
    """
    Turns a NumPy array into a GDAL dataset or exported as a raster using a reference raster.

    Parameters
    ----------
    array : np.ndarray
        The numpy array to convert.

    reference : str or gdal.Dataset
        The reference raster to use for the output.

    out_path : path, optional
        The destination to save to. Default: None.

    set_nodata : str, float, int, or bool. Optional
        Can be set to:
            - "arr": The nodata value will be the same as the NumPy array.
            - "ref": The nodata value will be the same as the reference raster.
            - value: The nodata value will be the value provided. Default: "arr".
            - None/False: The nodata value will not be set.

    allow_mismatches : bool, optional
        If True, the array can have a different shape than the reference raster.
        Default: False.

    pixel_offsets : list or tuple, optional
        If provided, the array will be written to the reference raster at the
        specified pixel offsets. The list should be in the format [x_offset, y_offset, x_size, y_size].
        Default: None.

    bbox : list, optional
        If provided, the array will be written to the reference raster at the specified
        bounding box. The list should be in the format [min_x, min_y, max_x, max_y]. Default: None.

    bbox_srs : str or osr.SpatialReference, optional
        The spatial reference of the bounding box. If None, the spatial reference
        of the raster is used. Default: None.

    channel_last : bool, optional
        If True, the array is in the channel-last format. If False, the array is in the
        channel-first format. Default: True.

    overwrite : bool, optional
        If the file exists, should it be overwritten? Default: True.

    creation_options : list, optional
        List of GDAL creation options. Default: ["TILED=YES", "NUM_THREADS=ALL_CPUS",
        "BIGTIFF=YES", "COMPRESS=LZW"].

    Returns
    -------
    str
        The file path to the newly created raster(s).

    Examples
    --------
    ```python
    >>> # Create a raster from a numpy array.
    >>> import buteo as beo
    >>> 
    >>> raster = "/path/to/raster/raster.tif"
    >>> 
    >>> array = beo.raster_to_array(raster)
    >>> array = array ** 2
    >>> 
    >>> out_path = beo.array_to_raster(
    ...     array,
    ...     reference=raster,
    ...     out_path="/path/to/new/new_raster.tif"
    ... )
    >>> 
    >>> out_path
    >>> "/path/to/new/new_raster.tif"
    ```
    """
    utils_base._type_check(array, [np.ndarray], "array")
    utils_base._type_check(reference, [str, gdal.Dataset], "reference")
    utils_base._type_check(out_path, [str, None], "out_path")
    utils_base._type_check(set_nodata, [float, int, str, bool, None], "set_nodata")
    utils_base._type_check(allow_mismatches, [bool], "allow_mismatches")
    utils_base._type_check(pixel_offsets, [list, tuple, None], "pixel_offsets")
    utils_base._type_check(bbox, [list, None], "bbox")
    utils_base._type_check(bbox_srs, [str, osr.SpatialReference, None], "bbox_srs")
    utils_base._type_check(channel_last, [bool], "channel_last")
    utils_base._type_check(overwrite, [bool], "overwrite")
    utils_base._type_check(creation_options, [list, None], "creation_options")

    if isinstance(set_nodata, str) and set_nodata not in ["arr", "ref"]:
        raise ValueError("set_nodata must be either 'arr' or 'ref'.")

    if array.ndim not in [2, 3]:
        raise ValueError("Array must be 2D or 3D.")

    assert utils_gdal._check_is_raster(reference), "Reference is not a raster."

    if out_path is None:
        out_path = utils_path._get_temp_filepath(name="array_to_raster", ext="tif")
    else:
        assert utils_path._check_is_valid_output_filepath(out_path, overwrite=overwrite), (
            f"Output path {out_path} is not valid or already exists. "
            "Set overwrite to True to overwrite existing files."
        )

    # Always output tif
    out_path = utils_path._get_augmented_path(out_path, change_ext="tif")

    if pixel_offsets is not None and bbox is not None:
        raise ValueError("Cannot specify both pixel offsets and bounding box.")

    metadata_ref = core_raster._get_basic_metadata_raster(reference)

    if pixel_offsets is None:
        pixel_offsets = [0, 0, metadata_ref["width"], metadata_ref["height"]]

    assert len(pixel_offsets) == 4 and all([isinstance(val, int) for val in pixel_offsets]), (
        "Pixel offsets must be a list of 4 integers."
    )

    if bbox is not None:
        assert len(bbox) == 4 and all([isinstance(val, (float, int)) for val in bbox]), (
            "Bounding box must be a list of 4 floats or integers."
        )
        assert utils_bbox._check_is_valid_bbox(bbox), "Bounding box is not valid."

        if bbox_srs is not None:
            bbox = utils_projection.reproject_bbox(bbox, bbox_srs, metadata_ref["projection_wkt"])

        pixel_offsets = utils_bbox._get_pixel_offsets(metadata_ref["geotransform"], bbox)

    # Handle nodata
    destination_nodata = False
    destination_nodata_value = 0.0
    if set_nodata == "arr":
        if np.ma.isMaskedArray(array):
            destination_nodata = True
            destination_nodata = array.fill_value
        else:
            destination_nodata = False
            destination_nodata = None
    elif set_nodata == "ref":
        destination_nodata = metadata_ref["nodata"]
        destination_nodata_value = metadata_ref["nodata_value"]
    elif set_nodata is None or set_nodata is False:
        destination_nodata = False
        destination_nodata_value = None
    else:
        destination_nodata = True
        destination_nodata_value = set_nodata

        if np.ma.isMaskedArray(array):
            array = np.ma.getdata(array.filled(destination_nodata_value))

    # Swap to channel_last for GDAL.
    if not channel_last:
        array = np.transpose(array, (1, 2, 0))

    x_start, y_start, x_size, y_size = pixel_offsets

    if x_size <= 0 or y_size <= 0:
        raise ValueError("Pixel offsets are invalid. Sizes must be greater than 0.")

    if array.shape[1] != x_size or array.shape[0] != y_size:
        raise ValueError("Array shape does not match pixel offsets.")

    bands = array.shape[2] if array.ndim == 3 else 1

    if not allow_mismatches and y_start + y_size > metadata_ref["height"]:
        raise ValueError("Array is too large for reference raster.")

    if not allow_mismatches and x_start + x_size > metadata_ref["width"]:
        raise ValueError("Array is too large for reference raster.")

    if array.ndim == 3:
        array = array[:y_size, :x_size:, :]
    else:
        array = array[:y_size, :x_size]

    # Create output raster
    driver = utils_gdal._get_default_driver_raster()

    destination_transform = (
        metadata_ref["geotransform"][0] + (x_start * metadata_ref["pixel_width"]),
        metadata_ref["geotransform"][1],
        metadata_ref["geotransform"][2],
        metadata_ref["geotransform"][3] - (y_start * metadata_ref["pixel_height"]),
        metadata_ref["geotransform"][4],
        metadata_ref["geotransform"][5],
    )

    utils_path._delete_if_required(out_path, overwrite)

    destination = driver.Create(
        out_path,
        x_size,
        y_size,
        bands,
        utils_translate._translate_dtype_numpy_to_gdal(array.dtype),
        utils_gdal._get_default_creation_options(creation_options),
    )

    if destination is None:
        raise RuntimeError(f"Could not create output raster: {out_path}")

    destination.SetGeoTransform(destination_transform)
    destination.SetProjection(metadata_ref["projection_wkt"])

    for band_idx in range(bands):
        band = destination.GetRasterBand(band_idx + 1)
        band.SetColorInterpretation(gdal.GCI_Undefined)

        if bands > 1 or array.ndim == 3:
            band.WriteArray(array[:, :, band_idx])
        else:
            band.WriteArray(array)

        if destination_nodata:
            if utils_base._check_variable_is_int(destination_nodata_value):
                band.SetNoDataValue(int(destination_nodata_value))
            elif utils_base._check_variable_is_float(destination_nodata_value):
                band.SetNoDataValue(float(destination_nodata_value))
            else:
                band.SetNoDataValue(np.nan)

    destination.FlushCache()
    destination = None

    return out_path


def save_dataset_to_disk(
    dataset: Union[gdal.Dataset, ogr.DataSource, str, List[Union[gdal.Dataset, ogr.DataSource, str]]],
    out_path: Union[str, List[str]],
    prefix: str = "",
    suffix: str = "",
    add_uuid: bool = False,
    add_timestamp: bool = False,
    overwrite: bool = True,
    creation_options: Optional[List[str]] = None,
) -> Union[str, List[str]]:
    """
    Writes a dataset to disk. Can be a raster or a vector.

    Parameters
    ----------
    dataset : Union[gdal.Dataset, ogr.DataSource, str]
        The dataset to write to disk.

    out_path : Union[str, List[str]]
        The output path or list of output paths.
    
    prefix : str, optional
        A prefix to add to the output path. Default: "".

    suffix : str, optional
        A suffix to add to the output path. Default: "".

    add_uuid : bool, optional
        If True, a UUID will be added to the output path. Default: False.

    add_timestamp : bool, optional
        If True, a timestamp will be added to the output path. Default: False.

    overwrite : bool, optional
        If True, the output will be overwritten if it already exists. Default: True.

    creation_options : Optional[List[str]], optional
        A list of creation options. Default: None.

    Returns
    -------
    Union[str, List[str]]
        The output path or list of output paths.
    """
    input_is_list = isinstance(dataset, list)

    input_data = utils_io._get_input_paths(dataset, "mixed")
    output_paths = utils_io._get_output_paths(
        input_data,
        out_path,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
        add_timestamp=add_timestamp,
        overwrite=overwrite,
    )
    creation_options = utils_gdal._get_default_creation_options(creation_options)

    for idx, dataset in enumerate(input_data):
        driver = None

        # Raster
        if utils_gdal._check_is_raster(dataset):
            driver_name = utils_gdal._get_driver_name_from_path(dataset)
            driver = gdal.GetDriverByName(driver_name)
            src_ds = gdal.Open(dataset)
            utils_path._delete_if_required(output_paths[idx], overwrite)
            driver.CreateCopy(output_paths[idx], src_ds, options=creation_options)
            src_ds = None

        # Vector
        elif utils_gdal._check_is_vector(dataset):
            driver_name = utils_gdal._get_driver_name_from_path(dataset)
            driver = ogr.GetDriverByName(driver_name)
            src_ds = ogr.Open(dataset)
            utils_path._delete_if_required(output_paths[idx], overwrite)
            driver.CopyDataSource(src_ds, output_paths[idx])
            src_ds = None

        else:
            raise RuntimeError(f"Invalid dataset type: {dataset}")

        if driver is None:
            raise RuntimeError("Could not get driver for output dataset.")

    if input_is_list:
        return output_paths

    return output_paths[0]


class raster_to_array_chunks:
    """
    A class for reading raster data in chunks. The array will be split into x and y
    amount of chunks in the x and y directions. The output will be the read array
    and the offsets of the chunk in the raster. The offset can be used to reconstitute
    the array into the original raster or a new raster representing the chunk,
    using the :func:`array_to_raster` function.

    Parameters
    ----------
    raster : Union[gdal.Dataset, str, List[Union[str, gdal.Dataset]]]
        The raster to read.

    chunks : int
        The number of chunks to read. The area is chunked in way that ensures
        that the chunks are as square as possible. Default: 1.

    chunk_size : list or tuple, optional
        The raster can be split into chunks of a fixed size, 
        instead of splitting into a fixed number of chunks.

        The list should be in the format [x_size, y_size].
        If this is provided, the chunks parameter is ignored. Default: None.

    overlap : int, optional
        The number of pixels to overlap. Default: 0.

    bands : list or str or int, optional
        The bands to read. Can be "all", an int, or a list of integers, or a single
        integer. Please note that bands are 1-indexed. Default: "all".

    filled : bool, optional
        Whether to fill masked values. Default: False.

    fill_value : int or float, optional
        The value to fill masked values with. Default: None.

    skip : int, optional
        The number of chunks to skip when iterating. Default: 0.

    border_strategy : int, optional
        The border strategy to use when splitting the raster into chunks.
        border_strategy ignored when chunk_size and overlaps are provided.
        Only applied when chunk_size is provided. Can be 1 or 2. Default: 1.
        1. Ignore the border chunks if they do not fit the chunk size.
        2. Oversample the border chunks to fit the chunk size.
        3. Shrink the last chunk to fit the image size. (Creates uneven chunks.)

    cast : type or str, optional
        The data type to cast the output to. Default: None.

    Returns
    -------
    generator
        A generator that yields the raster data in chunks and the offsets of the chunk
        in the raster in a tuple.

    Examples
    --------
    ```python
    >>> # Read a raster into array via chunks.
    >>> import buteo as beo
    >>> 
    >>> raster = "/path/to/raster/raster.tif"
    >>> 
    >>> shape = beo.raster_to_metadata(raster)["shape"]
    >>> shape
    >>> (100, 100)
    >>> 
    >>> for chunk, offsets in beo.raster_to_array_chunks(raster, chunks=4):
    >>>     print(chunk.shape, offsets)
    >>>     (25, 25), (0, 0, 25, 25)
    ```
    """

    def __init__(
        self,
        raster: Union[gdal.Dataset, str, List[Union[str, gdal.Dataset]]],
        chunks: Optional[int] = 1,
        chunk_size: Optional[Union[List[int], Tuple[int, int]]] = None,
        *,
        overlap: int = 0,
        bands: Union[List[int], str, int] = 'all',
        filled: bool = False,
        fill_value: Optional[Union[int, float]] = None,
        skip: int = 0,
        border_strategy: int = 1,
        cast: Optional[Union[np.dtype, str]] = None,
        channel_last: bool = True,
    ):
        self.raster = raster
        self.chunks = chunks
        self.chunk_size = chunk_size
        self.overlap = overlap
        self.bands = bands
        self.filled = filled
        self.fill_value = fill_value
        self.skip = skip
        self.border_strategy = border_strategy
        self.cast = cast
        self.current_chunk = 0
        self.channel_last = channel_last

        self.shape = core_raster._get_basic_metadata_raster(self.raster)["shape"]

        assert self.chunks > 0, "The number of chunks must be greater than 0."
        assert self.overlap >= 0, "The overlap must be greater than or equal to 0."
        assert self.chunks <= self.shape[1], "The number of chunks must be less than or equal to the number of columns in the raster."
        assert self.chunks <= self.shape[0], "The number of chunks must be less than or equal to the number of rows in the raster."
        assert self.border_strategy in [1, 2, 3], "The border strategy must be 1, 2, or 3."

        if self.chunk_size is not None:
            assert isinstance(self.chunk_size, (list, tuple)), "Chunk size must be a list or tuple."
            assert len(self.chunk_size) == 2, "Chunk size must be a list or tuple of length 2."
            assert all([isinstance(val, int) for val in self.chunk_size]), "Chunk size must be a list or tuple of integers."

            self.offsets = core_offsets._get_chunk_offsets_fixed_size(
                self.shape,
                self.chunk_size[0],
                self.chunk_size[1],
                self.border_strategy,
                self.overlap,
                channel_last=channel_last,
            )

        else:
            self.offsets = core_offsets._get_chunk_offsets(
                self.shape,
                self.chunks,
                self.overlap,
                channel_last=channel_last,
            )

        self.total_chunks = len(self.offsets)

    def __iter__(self):
        self.current_chunk = 0
        return self

    def __next__(self) -> Tuple[np.ndarray, List[int]]:
        if self.current_chunk >= self.total_chunks:
            raise StopIteration

        offset = self.offsets[self.current_chunk]
        self.current_chunk += 1 + self.skip

        return (
            raster_to_array(
                self.raster,
                bands=self.bands,
                filled=self.filled,
                fill_value=self.fill_value,
                pixel_offsets=offset,
                cast=self.cast,
                channel_last=self.channel_last,
            ),
            offset,
        )

    def __len__(self):
        return self.total_chunks


def raster_create_empty(
    out_path: Union[str, None] = None,
    *,
    width: int = 100,
    height: int = 100,
    pixel_size: Union[Union[float, int], List[Union[float, int]]] = 10.0,
    bands: int = 1,
    dtype: Union[str, int, np.dtype] = "uint8",
    x_min: Union[float, int] = 0.0,
    y_max: Union[float, int] = 0.0,
    nodata_value: Union[float, int, None] = None,
    fill_value: Union[float, int, None] = 0.0,
    projection: Union[int, str, gdal.Dataset, ogr.DataSource, osr.SpatialReference] = "EPSG:3857",
    creation_options: Union[List[str], None] = None,
    overwrite: bool = True,
) -> str:
    """
    Create an empty raster.

    Parameters
    ----------
    out_path : str, optional
        The output path. If None, a temporary file will be created.

    width : int, optional
        The width of the raster in pixels. Default: 100.

    height : int, optional
        The height of the raster in pixels. Default: 100.

    pixel_size : int or float or list or tuple, optional
        The pixel size in units of the projection. Default: 10.0.

    bands : int, optional
        The number of bands in the raster. Default: 1.

    dtype : str, int, np.dtype, optional
        The data type of the raster. Default: "uint8".

    x_min : int or float, optional
        The x coordinate of the top left corner of the raster. Default: 0.0.

    y_max : int or float, optional
        The y coordinate of the top left corner of the raster. Default: 0.0.

    nodata_value : int or float or None, optional
        The nodata value of the raster. Default: None.

    fill_value : int or float or None, optional
        The fill value of the raster. Default: 0.0.

    projection : int or str or gdal.Dataset or ogr.DataSource or osr.SpatialReference, optional
        The projection of the raster. Default: "EPSG:3857".

    creation_options : list or None, optional
        A list of creation options. Default: None.

    overwrite : bool, optional
        If True, overwrite the output file if it exists. Default: True.

    Returns
    -------
    str
        The path to the output raster.
    """
    utils_base._type_check(out_path, [str, type(None)], "out_path")
    utils_base._type_check(width, [int], "width")
    utils_base._type_check(height, [int], "height")
    utils_base._type_check(pixel_size, [int, float, list, tuple], "pixel_size")
    utils_base._type_check(bands, [int], "bands")
    utils_base._type_check(dtype, [str, int, np.dtype, type(np.int8), None], "dtype")
    utils_base._type_check(x_min, [int, float], "x_min")
    utils_base._type_check(y_max, [int, float], "y_max")
    utils_base._type_check(nodata_value, [int, float, type(None)], "nodata_value")
    utils_base._type_check(fill_value, [int, float, type(None)], "fill_value")
    utils_base._type_check(projection, [int, str, gdal.Dataset, ogr.DataSource, osr.SpatialReference], "projection")
    utils_base._type_check(creation_options, [list, type(None)], "creation_options")
    utils_base._type_check(overwrite, bool, "overwrite")

    if out_path is not None and not utils_path._check_is_valid_output_filepath(out_path, overwrite=overwrite):
        raise ValueError(
            f"Output path {out_path} is not valid or already exists. "
            "Set overwrite to True to overwrite existing files."
        )

    if out_path is None:
        out_path = utils_path._get_temp_filepath(name="raster_create_empty", ext="tif", add_timestamp=True)

    driver_name = utils_gdal._get_driver_name_from_path(out_path)
    driver = gdal.GetDriverByName(driver_name)

    utils_path._delete_if_required(out_path, overwrite)

    destination = driver.Create(
        out_path,
        width,
        height,
        bands,
        utils_translate._translate_dtype_numpy_to_gdal(utils_translate._parse_dtype(dtype)),
        utils_gdal._get_default_creation_options(creation_options),
    )

    parsed_projection = utils_projection.parse_projection(projection, return_wkt=True)
    destination.SetProjection(parsed_projection)

    pixel_width = pixel_size if isinstance(pixel_size, (int,  float)) else pixel_size[0]
    pixel_height = pixel_size if isinstance(pixel_size, (int,  float)) else pixel_size[1]

    transform = [x_min, pixel_width, 0, y_max, 0, -pixel_height] # negative for north-up

    destination.SetGeoTransform(transform)

    if nodata_value is not None:
        for band in range(1, bands + 1):
            band_obj = destination.GetRasterBand(band)

            if nodata_value is not None:
                band_obj.SetNoDataValue(nodata_value)

            if fill_value is not None:
                band_obj.Fill(fill_value)

    destination.FlushCache()
    destination = None

    return out_path


def raster_create_from_array(
    arr: np.ndarray,
    out_path: str = None,
    pixel_size: Union[Union[float, int], List[Union[float, int]]] = 1.0,
    x_min: Union[float, int] = 0.0,
    y_max: Union[float, int] = 0.0,
    projection: Union[int, str, gdal.Dataset, ogr.DataSource, osr.SpatialReference] = "EPSG:3857",
    channel_last: bool = True,
    overwrite: bool = True,
    creation_options: Union[List[str], None] = None,
) -> str:
    """ Create a raster from a numpy array.

    Parameters
    ----------
    arr : np.ndarray
        The array to convert to a raster.

    out_path : str, optional
        The output path. If None, a temporary file will be created.

    pixel_size : int or float or list or tuple, optional
        The pixel size of the output raster. Default: 10.0.

    x_min : int or float, optional
        The x coordinate of the top left corner of the output raster. Default: 0.0.

    y_max : int or float, optional
        The y coordinate of the top left corner of the output raster. Default: 0.0.

    projection : int or str or gdal.Dataset or ogr.DataSource or osr.SpatialReference, optional
        The projection of the output raster. Default: "EPSG:3857".

    creation_options : list or None, optional
        The creation options for the output raster. Default: None.

    overwrite : bool, optional
        If True, the output raster will be overwritten if it already exists. Default: True.

    Returns
    -------
    str
        The path to the output raster.
    """
    utils_base._type_check(arr, [np.ndarray, np.ma.MaskedArray], "arr")
    utils_base._type_check(out_path, [str, None], "out_path")
    utils_base._type_check(pixel_size, [int, float, [int, float], tuple], "pixel_size")
    utils_base._type_check(x_min, [int, float], "x_min")
    utils_base._type_check(y_max, [int, float], "y_max")
    utils_base._type_check(projection, [int, str, gdal.Dataset, ogr.DataSource, osr.SpatialReference], "projection")
    utils_base._type_check(creation_options, [[str], None], "creation_options")
    utils_base._type_check(overwrite, [bool], "overwrite")

    assert arr.ndim in [2, 3], "Array must be 2 or 3 dimensional (3rd dimension considered bands.)"

    if arr.ndim == 2:
        arr = arr[:, :, np.newaxis]

    if not channel_last:
        arr = np.transpose(arr, (2, 0, 1))

    if out_path is not None and not utils_path._check_is_valid_output_filepath(out_path, overwrite=overwrite):
        raise ValueError(
            f"Output path {out_path} is not valid or already exists. "
            "Set overwrite to True to overwrite existing files."
        )

    if out_path is None:
        out_path = utils_path._get_temp_filepath(name="raster_create_from_array", ext="tif", add_timestamp=True)

    driver_name = utils_gdal._get_driver_name_from_path(out_path)
    driver = gdal.GetDriverByName(driver_name)

    utils_path._delete_if_required(out_path, overwrite)

    height, width, bands = arr.shape

    destination = driver.Create(
        out_path,
        width,
        height,
        bands,
        utils_translate._translate_dtype_numpy_to_gdal(arr.dtype),
        utils_gdal._get_default_creation_options(creation_options),
    )

    parsed_projection = utils_projection.parse_projection(projection, return_wkt=True)

    destination.SetProjection(parsed_projection)

    pixel_width = pixel_size if isinstance(pixel_size, (int,  float)) else pixel_size[0]
    pixel_height = pixel_size if isinstance(pixel_size, (int,  float)) else pixel_size[1]

    transform = [x_min, pixel_width, 0, y_max, 0, -pixel_height] # negative for north-up

    destination.SetGeoTransform(transform)

    nodata = None
    nodata_value = 0.0

    if isinstance(arr, np.ma.MaskedArray):
        nodata = True
        nodata_value = arr.fill_value
        arr = np.ma.getdata(arr.filled(nodata_value))

    for idx in range(0, bands):
        dst_band = destination.GetRasterBand(idx + 1)
        dst_band.WriteArray(arr[:, :, idx])

        if nodata:
            dst_band.SetNoDataValue(nodata_value)

    return out_path


def raster_create_copy(
    raster: Union[str, gdal.Dataset],
    out_path: Optional[str] = None,
    overwrite: bool = True,
) -> str:
    """
    Create a copy of a raster.

    Parameters
    ----------
    raster : str or gdal.Dataset
        The raster to copy.

    out_path : str, optional
        The output path. If None, a temporary file will be created.

    Returns
    -------
    str
        The path to the output raster.
    """
    assert utils_gdal._check_is_raster(raster), "Raster is not valid."

    if out_path is None:
        out_path = utils_path._get_temp_filepath(
            utils_gdal._get_path_from_dataset(raster),
            ext="tif",
        )
    else:
        assert utils_path._check_is_valid_output_filepath(out_path, overwrite=overwrite), (
            f"Output path {out_path} is not valid or already exists. "
        )

    utils_path._delete_if_required(out_path, overwrite)

    driver_name = utils_gdal._get_driver_name_from_path(out_path)
    driver = gdal.GetDriverByName(driver_name)

    src_ds = core_raster._raster_open(raster)
    dst_ds = driver.CreateCopy(out_path, src_ds) # pylint: disable=unused-variable
    dst_ds = None
    src_ds = None

    return out_path



"""
Tools for working with sentinel data
"""
from .augmentation import *
from .augmentation_funcs import *
from .augmentation_utils import *
from .scalers import *
from .selection import *
from .encoding import *
from .masking_funcs import *
from .masking import *
from .spatial_label_smoothing import *


"""
Create offsets to read rasters in chunks.
"""

# Standard library
from typing import List, Tuple
from math import ceil



def _get_chunk_offsets(
    image_shape: Tuple[int, int],
    num_chunks: int,
    overlap: int = 0,
    channel_last: bool = True,
) -> List[Tuple[int, int, int, int]]:
    """
    Calculate chunk offsets for dividing an image into a specified number of chunks with minimal circumference.

    The function finds the optimal configuration of chunks to minimize the circumference and ensure the whole image
    is captured.

    Parameters
    ----------
    image_shape : Tuple[int, int]
        A tuple containing the height and width of the image. (Height, Width)

    num_chunks : int
        The number of chunks to divide the image into.

    overlap : int, optional
        The amount of overlap to apply to each chunk, in pixels. Default: 0.

    channel_last : bool, optional
        Whether the image has the channels as the last dimension. Default: True.

    Returns
    -------
    List[Tuple[int, int, int, int]]
        A list of tuples, each containing the chunk offsets and dimensions in the format: `(x_start, y_start, x_pixels, y_pixels)`.

    Raises
    ------
    ValueError
        If the number of chunks is too high for the given image size.
    """
    height, width = image_shape[:2] if channel_last else image_shape[1:]

    best_factors = None
    min_aspect_ratio_diff = float("inf")

    for i in range(1, num_chunks + 1):
        if num_chunks % i == 0:
            num_h_chunks = i
            num_w_chunks = num_chunks // i

            chunk_height = height // num_h_chunks
            chunk_width = width // num_w_chunks

            aspect_ratio = chunk_width / chunk_height
            aspect_ratio_diff = abs(aspect_ratio - 1)

            if aspect_ratio_diff < min_aspect_ratio_diff:
                min_aspect_ratio_diff = aspect_ratio_diff
                best_factors = (num_h_chunks, num_w_chunks)

    num_h_chunks, num_w_chunks = best_factors

    chunk_offsets = []

    for h in range(num_h_chunks):
        for w in range(num_w_chunks):
            h_start = h * (height // num_h_chunks)
            w_start = w * (width // num_w_chunks)

            h_end = height if h == num_h_chunks - 1 else (h + 1) * (height // num_h_chunks)
            w_end = width if w == num_w_chunks - 1 else (w + 1) * (width // num_w_chunks)

            x_pixels = w_end - w_start
            y_pixels = h_end - h_start

            chunk_offsets.append((w_start, h_start, x_pixels, y_pixels))

    if overlap > 0:
        overlap_half = ceil(overlap / 2)

        new_offsets = []
        for start_x, start_y, size_x, size_y in chunk_offsets:
            new_start_x = max(0, start_x - overlap_half)
            new_start_y = max(0, start_y - overlap_half)

            new_size_x = size_x + overlap_half
            new_size_y = size_y + overlap_half

            # If we are over the adjust, bring it back.
            if new_size_x + new_start_x > width:
                new_size_x = new_size_x - ((new_size_x + new_start_x) - width)

            if new_size_y + new_start_y > height:
                new_size_y = new_size_y - ((new_size_y + new_start_y) - height)

            new_offsets.append((
                new_start_x, new_start_y, new_size_x, new_size_y,
            ))

        return new_offsets

    return chunk_offsets


def _get_chunk_offsets_fixed_size(
    image_shape,
    chunk_size_x: int,
    chunk_size_y: int,
    border_strategy: int = 1,
    overlap: int = 0,
    *,
    channel_last: bool = True,
) -> List[Tuple[int, int, int, int]]:
    """
    Get chunk offsets for a fixed chunk size.

    Parameters
    ----------
    image_shape : Tuple[int, int]
        A tuple containing the height and width of the image. (Height, Width)

    chunk_size_x : int
        The size of the chunks in the x-direction.

    chunk_size_y : int
        The size of the chunks in the y-direction.

    border_strategy : int, optional
        The border strategy to use when splitting the raster into chunks.
        border_strategy ignored when chunk_size and overlaps are provided.
        Only applied when chunk_size is provided. Can be 1 or 2. Default: 1.
        1. Ignore the border chunks if they do not fit the chunk size.
        2. Oversample the border chunks to fit the chunk size.
        3. Shrink the last chunk to fit the image size. (Creates uneven chunks.)

    overlap : int, optional
        The amount of overlap to apply to each chunk, in pixels. Default: 0.

    channel_last : bool, optional
        Whether the image has the channels as the last dimension. Default: True.

    Returns
    -------
    List[Tuple[int, int, int, int]]
        A list of tuples, each containing the chunk offsets and dimensions in the format: `(x_start, y_start, x_pixels, y_pixels)`.
    """
    assert isinstance(image_shape, (list, tuple)), "image_shape must be a list or tuple."
    assert isinstance(chunk_size_x, int), "chunk_size_x must be an integer."
    assert isinstance(chunk_size_y, int), "chunk_size_y must be an integer."
    assert isinstance(border_strategy, int), "border_strategy must be an integer."
    assert isinstance(channel_last, bool), "channel_last must be a boolean."
    assert chunk_size_x > 0, "chunk_size_x must be > 0."
    assert border_strategy in [1, 2, 3], "border_strategy must be 1 or 2."
    height, width = image_shape[:2] if channel_last else image_shape[1:]

    num_chunks_x = width // chunk_size_x + (1 if width % chunk_size_x > 0 else 0)
    num_chunks_y = height // chunk_size_y + (1 if height % chunk_size_y > 0 else 0)

    chunk_offsets = []

    for y in range(num_chunks_y):
        for x in range(num_chunks_x):
            x_start = x * chunk_size_x + overlap
            y_start = y * chunk_size_y + overlap

            x_end = (x + 1) * chunk_size_x + overlap
            y_end = (y + 1) * chunk_size_y + overlap

            if x_end > width or y_end > height:
                if border_strategy == 1:
                    if x_end > width or y_end > height:
                        continue
                elif border_strategy == 2:
                    x_end = min(x_end, width)
                    y_end = min(y_end, height)

            x_pixels = x_end - x_start
            y_pixels = y_end - y_start

            if border_strategy in [1, 2] and (x_pixels < chunk_size_x or y_pixels < chunk_size_y):
                x_pixels = chunk_size_x
                y_pixels = chunk_size_y

            if x_start + x_pixels > width:
                if border_strategy == 1:
                    continue
                elif border_strategy == 2:
                    x_start = width - x_pixels
                elif border_strategy == 3:
                    x_pixels = width - x_start

            if y_start + y_pixels > height:
                if border_strategy == 1:
                    continue
                elif border_strategy == 2:
                    y_start = height - y_pixels
                elif border_strategy == 3:
                    y_pixels = height - y_start

            new_offset = (x_start, y_start, x_pixels, y_pixels)

            if border_strategy in [1, 2] and (new_offset[2] != chunk_size_x or new_offset[3] != chunk_size_y):
                raise RuntimeError("Parsing error in offsets.")

            if new_offset not in chunk_offsets:
                chunk_offsets.append(new_offset)

    if overlap > 0:
        num_chunks_x = width // chunk_size_x + (1 if width % chunk_size_x > 0 else 0)
        num_chunks_y = height // chunk_size_y + (1 if height % chunk_size_y > 0 else 0)

        for x in range(num_chunks_x):
            x_start = x * chunk_size_x
            x_end = (x + 1) * chunk_size_x

            x_pixels = chunk_size_x

            if x_start < x_pixels and border_strategy == 1:
                continue

            if x_end > width:
                if border_strategy == 1:
                    continue
                elif border_strategy == 2:
                    x_start = width - chunk_size_x
                elif border_strategy == 3:
                    x_pixels = width - x_start

            new_offset = (x_start, 0, x_pixels, chunk_size_y)

            if border_strategy in [1, 2] and (new_offset[2] != chunk_size_x or new_offset[3] != chunk_size_y):
                raise RuntimeError("Parsing error in offsets.")

            if new_offset not in chunk_offsets:
                chunk_offsets.append(new_offset)

        for y in range(num_chunks_y):
            y_start = y * chunk_size_y
            y_end = (y + 1) * chunk_size_y

            y_pixels = chunk_size_y

            if y_start < y_pixels and border_strategy == 1:
                continue

            if y_end > height:
                if border_strategy == 1:
                    continue
                elif border_strategy == 2:
                    y_start = height - chunk_size_y
                elif border_strategy == 3:
                    y_pixels = height - y_start

            new_offset = (0, y_start, chunk_size_x, y_pixels)

            if border_strategy in [1, 2] and (new_offset[2] != chunk_size_x or new_offset[3] != chunk_size_y):
                raise RuntimeError("Parsing error in offsets.")

            if new_offset not in chunk_offsets:
                chunk_offsets.append(new_offset)

    return chunk_offsets



"""
Create vector grids from references or hardcode
"""

# TODO: Implement grid functions
# TODO: create_grid, create_grid_points, create_grid_linestrings

def _vector_create_grid():
    """ Not implemented."""
    raise RuntimeError("Not implemented")



"""
### Dissolve vector geometries. ###

Dissolve vectors by attributes or geometry.
"""

# Standard library
import sys; sys.path.append("../../")
from typing import Union, Optional, List

# External
from osgeo import ogr

# Internal
from buteo.utils import (
    utils_io,
    utils_base,
    utils_gdal,
    utils_path,
)
from buteo.vector import core_vector
from buteo.vector.metadata import _vector_to_metadata




def _vector_dissolve(
    vector: Union[str, ogr.DataSource],
    attribute: Optional[str] = None,
    out_path: Optional[str] = None,
    overwrite: bool = True,
    add_index: bool = True,
    process_layer: int = -1,
) -> Union[str, ogr.DataSource]:
    """ Internal. """
    assert utils_gdal._check_is_vector(vector), "Invalid input vector"

    vector_list = utils_base._get_variable_as_list(vector)

    if out_path is None:
        out_path = utils_path._get_temp_filepath("dissolve.shp", suffix="_dissolve")

    assert utils_path._check_is_valid_output_filepath(out_path, overwrite=overwrite), "Invalid output path"

    out_format = utils_gdal._get_vector_driver_name_from_path(out_path)

    driver = ogr.GetDriverByName(out_format)

    ref = core_vector._vector_open(vector_list[0])
    metadata = _vector_to_metadata(ref)

    layers = []

    if process_layer == -1:
        for index in range(len(metadata["layers"])):
            layers.append(
                {
                    "name": metadata["layers"][index]["layer_name"],
                    "geom": metadata["layers"][index]["column_geom"],
                    "fields": metadata["layers"][index]["field_names"],
                }
            )
    else:
        layers.append(
            {
                "name": metadata["layers"][process_layer]["layer_name"],
                "geom": metadata["layers"][process_layer]["column_geom"],
                "fields": metadata["layers"][process_layer]["field_names"],
            }
        )

    utils_path._delete_if_required(out_path, overwrite=overwrite)

    destination = driver.CreateDataSource(out_path)

    # Check if attribute table is valid
    for index in range(len(metadata["layers"])):
        layer = layers[index]
        if attribute is not None and attribute not in layer["fields"]:
            layer_fields = layer["fields"]
            raise ValueError(
                f"Invalid attribute for layer. Layers has the following fields: {layer_fields}"
            )

        geom_col = layer["geom"]
        name = layer["name"]

        try:
            sql = None
            if attribute is None:
                sql = f"SELECT ST_Union({geom_col}) AS geom FROM {name};"
            else:
                sql = f"SELECT {attribute}, ST_Union({geom_col}) AS geom FROM {name} GROUP BY {attribute};"

            result = ref.ExecuteSQL(sql, dialect="SQLITE")
        except:
            sql = None
            if attribute is None:
                sql = f"SELECT ST_Union(geometry) AS geom FROM {name};"
            else:
                sql = f"SELECT {attribute}, ST_Union(geometry) AS geom FROM {name} GROUP BY {attribute};"

            result = ref.ExecuteSQL(sql, dialect="SQLITE")

        destination.CopyLayer(result, name, ["OVERWRITE=YES"])

    if add_index:
        core_vector.vector_add_index(destination)

    destination.FlushCache()

    return out_path


def vector_dissolve(
    vector: Union[str, ogr.DataSource, List[Union[str, ogr.DataSource]]],
    attribute: Optional[str] = None,
    out_path: Optional[str] = None,
    add_index: bool = True,
    process_layer: int = -1,
    prefix: str = "",
    suffix: str = "",
    add_uuid: bool = False,
    overwrite: bool = True,
    allow_lists: bool = True,
):
    """
    Clips a vector to a geometry.

    Parameters
    ----------
    vector : Union[str, ogr.DataSource, List[Union[str, ogr.DataSource]]]
        The vector(s) to clip.

    attribute : Optional[str], optional
        The attribute to use for the dissolve, default: None
    
    out_path : Optional[str], optional
        The output path, default: None

    add_index : bool, optional
        Add a spatial index to the output, default: True

    process_layer : int, optional
        The layer to process, default: -1

    prefix : str, optional
        The prefix to add to the output path, default: ""

    suffix : str, optional
        The suffix to add to the output path, default: ""

    add_uuid : bool, optional
        Add a uuid to the output path, default: False

    overwrite : bool, optional
        Overwrite the output, default: True

    allow_lists : bool, optional
        Allow lists as input, default: True

    Returns
    -------
    Union[str, ogr.DataSource]
        The output path or ogr.DataSource
    """
    utils_base._type_check(vector, [ogr.DataSource, str, [str, ogr.DataSource]], "vector")
    utils_base._type_check(attribute, [str, None], "attribute")
    utils_base._type_check(out_path, [str, [str], None], "out_path")
    utils_base._type_check(add_index, [bool], "add_index")
    utils_base._type_check(process_layer, [int], "process_layer")
    utils_base._type_check(prefix, [str], "prefix")
    utils_base._type_check(suffix, [str], "suffix")
    utils_base._type_check(add_uuid, [bool], "add_uuid")
    utils_base._type_check(overwrite, [bool], "overwrite")
    utils_base._type_check(allow_lists, [bool], "allow_lists")

    if not allow_lists and isinstance(vector, list):
        raise ValueError("Lists are not allowed when allow_lists is False.")

    vector_list = utils_base._get_variable_as_list(vector)

    assert utils_gdal._check_is_vector_list(vector_list), f"Invalid input vector: {vector_list}"

    path_list = utils_io._get_output_paths(
        vector_list,
        out_path,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
    )

    assert utils_path._check_is_valid_output_path_list(path_list, overwrite=overwrite), f"Invalid output path generated. {path_list}"

    output = []
    for index, in_vector in enumerate(vector_list):
        output.append(
            _vector_dissolve(
                in_vector,
                attribute=attribute,
                out_path=path_list[index],
                overwrite=overwrite,
                add_index=add_index,
                process_layer=process_layer,
            )
        )

    if isinstance(vector, list):
        return output

    return output[0]



# Standard library
import sys; sys.path.append("../../")
from typing import Union, Optional, List

# External
import numpy as np
from osgeo import ogr

# Internal
from buteo.utils import (
    utils_io,
    utils_base,
    utils_gdal,
    utils_bbox,
    utils_path,
)
from buteo.vector.metadata import _vector_to_metadata
from buteo.vector.core_vector import _vector_open



def _vector_add_shapes_in_place(
    vector: Union[ogr.DataSource, str],
    shapes: Optional[List[str]] = None,
    prefix: str = "",
    verbose: bool = False,
) -> str:
    """ Internal. """
    assert isinstance(vector, (ogr.DataSource, str)), "vector must be a vector layer or path to one."
    assert isinstance(shapes, (list, tuple)) or shapes is None, "shapes must be a list of shapes."
    assert isinstance(prefix, str), "prefix must be a string."

    all_shapes = ["area", "perimeter", "ipq", "hull", "compactness", "centroid"]

    if shapes is None:
        shapes = all_shapes
    else:
        for shape in shapes:
            if shape not in all_shapes:
                raise ValueError(f"{shape} is not a valid shape.")

    datasource = _vector_open(vector)
    out_path = utils_gdal._get_path_from_dataset(datasource, dataset_type="vector")
    metadata = _vector_to_metadata(datasource)

    for index in range(metadata["layer_count"]):
        vector_current_fields = metadata["layers"][index]["field_names"]
        vector_layer = datasource.GetLayer(index)

        vector_layer.StartTransaction()

        # Add missing fields
        for attribute in shapes:
            if attribute == "centroid":
                if "centroid_x" not in vector_current_fields:
                    field_defn = ogr.FieldDefn(f"{prefix}centroid_x", ogr.OFTReal)
                    vector_layer.CreateField(field_defn)

                if "centroid_y" not in vector_current_fields:
                    field_defn = ogr.FieldDefn(f"{prefix}centroid_y", ogr.OFTReal)
                    vector_layer.CreateField(field_defn)

            elif attribute not in vector_current_fields:
                field_defn = ogr.FieldDefn(f"{prefix}{attribute}", ogr.OFTReal)
                vector_layer.CreateField(field_defn)

        vector_feature_count = vector_layer.GetFeatureCount()

        if verbose:
            utils_base.progress(0, vector_feature_count, name="shape")

        for i in range(vector_feature_count):
            vector_feature = vector_layer.GetNextFeature()

            try:
                vector_geom = vector_feature.GetGeometryRef()
            except RuntimeWarning:
                # vector_geom.Buffer(0)
                raise RuntimeWarning("Invalid geometry at : ", i) from None

            if vector_geom is None:
                raise RuntimeError("Invalid geometry. Could not fix.")

            centroid = vector_geom.Centroid()
            vector_area = vector_geom.GetArea()
            vector_perimeter = vector_geom.Boundary().Length()

            if "ipq" or "compact" in shapes:
                vector_ipq = 0
                if vector_perimeter != 0:
                    vector_ipq = (4 * np.pi * vector_area) / vector_perimeter ** 2

            if "centroid" in shapes:
                vector_feature.SetField(f"{prefix}centroid_x", centroid.GetX())
                vector_feature.SetField(f"{prefix}centroid_y", centroid.GetY())

            if "hull" in shapes or "compact" in shapes:
                vector_hull = vector_geom.ConvexHull()
                hull_area = vector_hull.GetArea()
                hull_peri = vector_hull.Boundary().Length()
                hull_ratio = float(vector_area) / float(hull_area)
                compactness = np.sqrt(float(hull_ratio) * float(vector_ipq))

            if "area" in shapes:
                vector_feature.SetField(f"{prefix}area", vector_area)
            if "perimeter" in shapes:
                vector_feature.SetField(f"{prefix}perimeter", vector_perimeter)
            if "ipq" in shapes:
                vector_feature.SetField(f"{prefix}ipq", vector_ipq)
            if "hull" in shapes:
                vector_feature.SetField(f"{prefix}hull_area", hull_area)
                vector_feature.SetField(f"{prefix}hull_peri", hull_peri)
                vector_feature.SetField(f"{prefix}hull_ratio", hull_ratio)
            if "compact" in shapes:
                vector_feature.SetField(f"{prefix}compact", compactness)

            vector_layer.SetFeature(vector_feature)

            if verbose:
                utils_base.progress(i, vector_feature_count, name="shape")

        vector_layer.CommitTransaction()

    return out_path


def vector_add_shapes_in_place(
    vector: Union[str, ogr.DataSource, List[Union[str, ogr.DataSource]]],
    shapes: Optional[List[str]] = None,
    prefix: str = "",
    allow_lists: bool = True,
    verbose: bool = False,
) -> Union[str, List[str]]:
    """
    Adds shape calculations to a vector such as area and perimeter.
    Can also add compactness measurements.

    Parameters
    ----------
    vector : Union[str, ogr.DataSource, List[str, ogr.DataSource]]
        Vector layer(s) or path(s) to vector layer(s).
    
    shapes : Optional[List[str]], optional
        The shapes to calculate. The following a possible:
            * Area          (In same unit as projection)
            * Perimeter     (In same unit as projection)
            * IPQ           (0-1) given as (4*Pi*Area)/(Perimeter ** 2)
            * Hull Area     (The area of the convex hull. Same unit as projection)
            * Compactness   (0-1) given as sqrt((area / hull_area) * ipq)
            * Centroid      (Coordinate of X and Y)
        Default: all shapes.
    
    prefix : str, optional
        Prefix to add to the field names. Default: "".

    allow_lists : bool, optional
        If True, will accept a list of vectors. If False, will raise an error if a list is passed. Default: True.

    verbose : bool, optional
        If True, will print progress. Default: False.

    Returns
    -------
    out_path : str
        Path to the output vector.
    """
    utils_base._type_check(vector, [str, ogr.DataSource, [str, ogr.DataSource]], "vector")
    utils_base._type_check(shapes, [[str], None], "shapes")

    if not allow_lists and isinstance(vector, list):
        raise ValueError("Lists of vectors are not supported when allow_list is False.")

    vector_list = utils_base._get_variable_as_list(vector)
    output = utils_gdal._get_path_from_dataset_list(vector_list)

    for in_vector in vector_list:
        output.append(_vector_add_shapes_in_place(
            in_vector,
            shapes=shapes,
            prefix=prefix,
            verbose=verbose,
        ))

    if isinstance(vector, list):
        return output

    return output[0]



"""
### Functions for changing the datatype of a raster. ###
"""

# Standard library
import sys; sys.path.append("../../")
from typing import Union, List, Optional

# External
from osgeo import gdal
import numpy as np

# Internal
from buteo.utils import (
    utils_io,
    utils_base,
    utils_path,
    utils_translate,
)
from buteo.raster import core_raster, core_raster_io



def raster_stack_list(
    rasters: List[Union[str, gdal.Dataset]],
    out_path: Optional[str] = None,
    *,
    overwrite: bool = True,
    dtype: Optional[str] = None,
    creation_options: Optional[List[str]] = None,
) -> Union[str, List[str]]:
    """
    Stacks a list of aligned rasters into a single raster file.

    Parameters
    ----------
    rasters : list
        List of rasters to stack. These rasters must be aligned.

    out_path : str or None, optional
        The destination to save the output raster. If not provided, a temporary file will be created. Default: None.

    overwrite : bool, optional
        If the file exists, should it be overwritten? Default: True.

    dtype : str, optional
        The data type of the output raster. If not provided, the data type of the input rasters will be used. Default: None.

    creation_options : list, optional
        A list of GDAL creation options for the output raster. Default is
        ["TILED=YES", "NUM_THREADS=ALL_CPUS", "BIGTIFF=YES", "COMPRESS=LZW"].

    Returns
    -------
    str or list
        The file path(s) to the newly created raster(s).
    """
    utils_base._type_check(rasters, [[str, gdal.Dataset]], "rasters")
    utils_base._type_check(out_path, [str, None], "out_path")
    utils_base._type_check(overwrite, [bool], "overwrite")
    utils_base._type_check(dtype, [str, None, np.dtype, type(np.uint8)], "dtype")
    utils_base._type_check(creation_options, [[str], None], "creation_options")

    # Get input raster file paths
    input_data = utils_io._get_input_paths(rasters, "raster")

    # Generate a temporary output file path if out_path is not provided
    if out_path is None:
        out_path = utils_path._get_temp_filepath(
            "temp_stack",
            ext="tif",
            add_uuid=True,
            add_timestamp=True,
        )
    else:
        assert utils_path._check_is_valid_output_filepath(out_path), "Invalid output path."

    # Check if input rasters are aligned
    assert core_raster.check_rasters_are_aligned(input_data), "Rasters are not aligned."

    # Read input rasters as NumPy arrays and stack them
    array = core_raster_io.raster_to_array(input_data, cast=dtype)

    # Write the stacked array as a raster file
    out_path = core_raster_io.array_to_raster(
        array,
        reference=input_data[0],
        out_path=out_path,
        overwrite=overwrite,
        creation_options=creation_options,
    )

    return out_path


def raster_stack_vrt_list(
    rasters: List[Union[str, gdal.Dataset]],
    out_path: Optional[str]=None,
    separate: bool = True,
    *,
    resample_alg: str = "nearest",
    nodata_src: Optional[float] = None,
    nodata_VRT: Optional[float] = None,
    nodata_hide: Optional[bool] = None,
    options: Optional[list] = None,
    overwrite: bool = True,
    reference: Optional[str] = None,
    creation_options: Optional[List[str]] = None,
) -> str:
    """
    Create a virtual raster (.vrt) by stacking a list of input rasters.

    The function accepts a list of rasters and creates a virtual raster stack
    by combining them. The rasters can be stacked by keeping their bands separate,
    or by merging the bands in the order of the input rasters.

    Parameters
    ----------
    rasters : List[Union[str, gdal.Dataset]]
        A list of input rasters as either file paths or GDAL datasets to be stacked.

    out_path : Optional[str], default=None
        The destination file path to save the output VRT raster. If not provided, a
        temporary file path will be generated.

    separate : bool, default=True
        If True, the raster bands will be kept separate and stacked in the order of
        the input rasters. If False, the raster bands will be merged, but all input
        rasters must have the same number of bands.

    resample_alg : str, default='nearest'
        The resampling algorithm to use when building the VRT. Accepts any algorithm
        supported by GDAL's BuildVRT function (e.g., 'nearest', 'bilinear', 'cubic').

    nodata_src : Optional[float], default=None
        The NoData value to use for the source rasters. If not provided, the NoData
        value from the input rasters will be used.

    nodata_VRT : Optional[float], default=None
        The NoData value to use for the output VRT raster. If not provided, the NoData
        value from the input rasters will be used.

    nodata_hide : Optional[bool], default=None
        If True, the NoData value will be hidden in the VRT. If not provided, the value
        will be determined by the input rasters.

    options : Optional[list], default=None
        A list of VRT options for GDAL. If not provided, default options will be used.

    overwrite : bool, default=True
        If True and the output file exists, it will be overwritten. If False and the
        output file exists, an error will be raised.

    reference : Optional[str], default=None
        A reference raster file path or GDAL dataset to use for aligning the stacked
        rasters. If not provided, the alignment of the input rasters will be used.

    creation_options : Optional[List[str]], default=None
        A list of GDAL creation options for the output VRT raster. If not provided,
        the default options will be used.

    Returns
    -------
    str
        The file path to the newly created VRT raster.
    """
    # Type checking for input parameters
    utils_base._type_check(rasters, [[str, gdal.Dataset]], "rasters")
    utils_base._type_check(out_path, [str, None], "out_path")
    utils_base._type_check(separate, [bool], "separate")
    utils_base._type_check(resample_alg, [str], "resample_alg")
    utils_base._type_check(options, [list, None], "options")
    utils_base._type_check(overwrite, [bool], "overwrite")
    utils_base._type_check(creation_options, [[str], None], "creation_options")

    # Check if all rasters have the same number of bands when separate is False
    if not separate:
        master_bands = 0

        for idx, raster_input in enumerate(rasters):
            if idx == 0:
                master_bands = core_raster._get_basic_metadata_raster(raster_input)["bands"]
            else:
                if master_bands != core_raster._get_basic_metadata_raster(raster_input)["bands"]:
                    raise ValueError("All rasters must have the same number of bands when separate is False.")

    # Generate a temporary output file path if out_path is not provided
    if out_path is None:
        out_path = utils_path._get_temp_filepath(
            "temp_stack",
            ext="vrt",
            add_uuid=True,
            add_timestamp=True,
        )

    # Translate the input resample algorithm to a GDAL-compatible format
    resample_algorithm = utils_translate._translate_resample_method(resample_alg)

    # Build VRT options based on whether a reference raster is provided or not
    if reference is not None:
        meta = core_raster._get_basic_metadata_raster(reference)
        options = gdal.BuildVRTOptions(
            resampleAlg=resample_algorithm,
            separate=separate,
            outputBounds=meta["bbox_gdal"],
            xRes=meta["pixel_width"],
            yRes=meta["pixel_height"],
            targetAlignedPixels=True,
            srcNodata=nodata_src,
            VRTNodata=nodata_VRT,
            hideNodata=nodata_hide,
        )
    else:
        options = gdal.BuildVRTOptions(
            resampleAlg=resample_algorithm,
            separate=separate,
            srcNodata=nodata_src,
            VRTNodata=nodata_VRT,
            hideNodata=nodata_hide,
        )

    # Create separate VRTs for each band if separate is True
    if separate:
        tmp_vrt_list = []

        for idx, raster in enumerate(rasters):
            bands_in_raster = core_raster._get_basic_metadata_raster(raster)["bands"]

            for band in range(bands_in_raster):
                tmp_vrt_path = utils_path._get_temp_filepath(
                    "temp_vrt",
                    ext="vrt",
                    add_uuid=True,
                    add_timestamp=True,
                )

                tmp_vrt_code = gdal.BuildVRT(
                    tmp_vrt_path,
                    raster,
                    options=gdal.BuildVRTOptions(
                        resampleAlg=resample_algorithm,
                        separate=True,
                        srcNodata=nodata_src,
                        VRTNodata=nodata_VRT,
                        hideNodata=nodata_hide,
                        bandList=[band + 1],
                    )
                )

                tmp_vrt_list.append(tmp_vrt_path)

                if tmp_vrt_code is None:
                    raise ValueError(f"Error while creating VRT from rasters: {rasters}")

                tmp_vrt_code = None

        vrt = gdal.BuildVRT(out_path, tmp_vrt_list, options=options)

        # Clean up temporary VRTs
        for tmp_vrt_path in tmp_vrt_list:
            gdal.Unlink(tmp_vrt_path)

    # Create a VRT by merging bands if separate is False
    else:
        vrt = gdal.BuildVRT(out_path, rasters, options=options)

    # Flush cache to ensure data is written to the output file
    vrt.FlushCache()

    # Check for errors in VRT creation
    if vrt is None:
        raise ValueError(f"Error while creating VRT from rasters: {rasters}")

    # Release the VRT object to avoid potential memory leaks
    vrt = None

    return out_path


# TODO: Mosaic raster(s)
# TODO: Use gdaltools?



"""
### GDAL Enum Functions ###

Functions to translate between **GDAL** and **NumPy** datatypes.
"""
# Standard Library
from typing import List, Tuple, Union, Dict, Type, Optional

# External
import numpy as np
from osgeo import gdal, gdal_array



def _get_available_drivers() -> List[Dict[str, str]]:
    """
    Returns a list of all available drivers.

    Returns
    -------
    List[Dict[str, str]]
        List of dicts containing available drivers. Each dict has the following keys:
        - 'short_name' (str): Driver short name (e.g. GTiff).
        - 'long_name' (str): Driver long name (e.g. GeoTiff).
        - 'extension' (str): Driver file extension (e.g. tif). Note: Can be an empty string.
    """
    raster_drivers = []
    vector_drivers = []

    for i in range(0, gdal.GetDriverCount()):
        driver = gdal.GetDriver(i)
        metadata = driver.GetMetadata_Dict()

        extensions = driver.GetMetadataItem(gdal.DMD_EXTENSIONS)
        if extensions is not None:
            extensions = extensions.split(" ")
        else:
            extensions = [None]

        for ext in extensions:
            info = {
                "short_name": driver.ShortName,
                "long_name": driver.LongName,
                "extension": ext,
            }

            if 'DCAP_RASTER' in metadata:
                raster_drivers.append(info)

            if 'DCAP_VECTOR' in metadata:
                vector_drivers.append(info)

    return (raster_drivers, vector_drivers)


def _get_valid_raster_driver_extensions() -> List[str]:
    """
    Returns a list of valid raster driver extensions.

    Returns
    -------
    List[str]
        List of valid raster driver extensions.
    """
    available_raster_drivers, _available_vector_drivers = _get_available_drivers()

    valid_raster_driver_extensions = []

    for driver in available_raster_drivers:
        if driver["extension"] != "" or len(driver["extension"]) > 0:

            if driver["extension"] == "gpkg": continue

            valid_raster_driver_extensions.append(driver["extension"])

    return valid_raster_driver_extensions


def _get_valid_vector_driver_extensions() -> List[str]:
    """
    Returns a list of valid vector driver extensions.

    Returns
    -------
    List[str]
        List of valid vector driver extensions.
    """
    _available_raster_drivers, available_vector_drivers = _get_available_drivers()

    valid_vector_driver_extensions = []

    for driver in available_vector_drivers:
        if driver["extension"] != "" or len(driver["extension"]) > 0:
            valid_vector_driver_extensions.append(driver["extension"])

    return valid_vector_driver_extensions


def _get_valid_driver_extensions() -> List[str]:
    """
    Returns a list of all valid driver extensions (**GDAL** + **OGR**).

    Returns
    -------
    List[str]
        List of all valid driver extensions.
    """
    available_raster_drivers, available_vector_drivers = _get_available_drivers()

    valid_driver_extensions = []

    for driver in available_raster_drivers + available_vector_drivers:
        if driver["extension"] != "" or len(driver["extension"]) > 0:
            valid_driver_extensions.append(driver["extension"])

    return valid_driver_extensions


def _check_is_valid_driver_extension(ext: str) -> bool:
    """
    Checks if a file extension is a valid GDAL or OGR driver extension.

    Parameters
    ----------
    ext : str
        The file extension.

    Returns
    -------
    bool
        True if valid, False otherwise.
    """
    assert isinstance(ext, str), "Extension must be a string."
    assert len(ext) > 0, "Extension must be a non-empty string."

    return ext in _get_valid_driver_extensions()


def _check_is_valid_raster_driver_extension(ext: str) -> bool:
    """
    Checks if a raster file extension is a valid GDAL driver extension.

    Parameters
    ----------
    ext : str
        The file extension.

    Returns
    -------
    bool
        True if valid, False otherwise.
    """
    assert isinstance(ext, str), "Extension must be a string."
    assert len(ext) > 0, "Extension must be a non-empty string."

    return ext in _get_valid_raster_driver_extensions()


def _check_is_valid_vector_driver_extension(ext: str) -> bool:
    """
    Checks if a vector file extension is a valid **OGR** driver extension.

    Parameters
    ----------
    ext : str
        The file extension.
    
    Returns
    -------
    bool
        True if valid, False otherwise.
    """
    assert isinstance(ext, str), "Extension must be a string."
    assert len(ext) > 0, "Extension must be a non-empty string."

    return ext in _get_valid_vector_driver_extensions()


def _get_raster_shortname_from_ext(ext: str) -> str:
    """
    Converts a raster file extension to a GDAL driver short name.

    Parameters
    ----------
    ext : str
        The file extension.

    Returns
    -------
    str
        The driver short name (e.g. GTiff).
    """
    assert _check_is_valid_raster_driver_extension(ext), f"Invalid extension: {ext}"

    raster_drivers, _vector_drivers = _get_available_drivers()

    for driver in raster_drivers:
        if driver["extension"] == ext:
            return driver["short_name"]

    raise RuntimeError(f"Invalid extension: {ext}")


def _get_vector_shortname_from_ext(ext: str) -> str:
    """
    Converts a vector file extension to an **OGR** driver short_name name.

    Parameters
    ----------
    ext : str
        The file extension.

    Returns
    -------
    str
        The driver short name (e.g. GPKG).
    """
    assert _check_is_valid_vector_driver_extension(ext), f"Invalid extension: {ext}"

    _raster_drivers, vector_drivers = _get_available_drivers()

    for driver in vector_drivers:
        if driver["extension"] == ext:
            return driver["short_name"]

    raise RuntimeError(f"Invalid extension: {ext}")


def _get_driver_shortname_from_ext(ext: str) -> str:
    """
    Converts a file extension to a driver short name for either OGR or GDAL.

    Parameters
    ----------
    ext : str
        The file extension.

    Returns
    -------
    str
        The driver short name (e.g. GPKG).
    """
    assert _check_is_valid_vector_driver_extension(ext) or _check_is_valid_raster_driver_extension(ext), f"Invalid extension: {ext}"

    raster_drivers, vector_drivers = _get_available_drivers()

    for driver in raster_drivers + vector_drivers:
        if driver["extension"] == ext:
            return driver["short_name"]

    raise RuntimeError(f"Invalid extension: {ext}")


def _translate_resample_method(method: str) -> int:
    """
    Translate a string of a resampling method to a GDAL integer (e.g. gdal.GRA_NearestNeighbour).

    Parameters
    ----------
    method : str
        The resampling method.

    Returns
    -------
    int
        The GDAL resampling method integer.
    """
    assert isinstance(method, str), "method must be a string."
    assert len(method) > 0, "method must be a non-empty string."

    methods = {
        "nearest": gdal.GRA_NearestNeighbour,
        "bilinear": gdal.GRA_Bilinear,
        "cubic": gdal.GRA_Cubic,
        "cubic_spline": gdal.GRA_CubicSpline,
        "cubicspline": gdal.GRA_CubicSpline,
        "lanczos": gdal.GRA_Lanczos,
        "average": gdal.GRA_Average,
        "mean": gdal.GRA_Average,
        "mode": gdal.GRA_Mode,
    }

    # Check GDAL version and add methods available in newer versions
    gdal_version = tuple(map(int, gdal.VersionInfo().split('.')))
    if gdal_version >= (3, 3, 0):
        extended_methods = {
            "max": gdal.GRA_Max,
            "maximum": gdal.GRA_Max,
            "min": gdal.GRA_Min,
            "minimum": gdal.GRA_Min,
            "median": gdal.GRA_Med,
            "med": gdal.GRA_Med,
            "q1": gdal.GRA_Q1,
            "Q1": gdal.GRA_Q1,
            "q3": gdal.GRA_Q3,
            "Q3": gdal.GRA_Q3,
            "sum": gdal.GRA_RMS,
            "rms": gdal.GRA_RMS,
            "RMS": gdal.GRA_RMS,
        }
        methods.update(extended_methods)
    else:
        print("Warning: Old version of GDAL running, only a subset of resampling methods are supported.")

    if method in methods:
        return methods[method]

    raise ValueError("Unknown resample method: " + method)


def _translate_dtype_gdal_to_numpy(gdal_datatype_int: int) -> np.dtype:
    """
    Translates the **GDAL** datatype integer into a **NumPy** datatype.

    Parameters
    ----------
    gdal_datatype_int : int
        The GDAL datatype integer.

    Returns
    -------
    np.dtype
        The **NumPy** datatype.
    """
    assert isinstance(gdal_datatype_int, int), f"gdal_datatype must be an integer. Received: {gdal_datatype_int}"

    dtype = np.dtype(gdal_array.GDALTypeCodeToNumericTypeCode(gdal_datatype_int))

    return dtype


def _translate_dtype_numpy_to_gdal(numpy_datatype: np.dtype) -> int:
    """
    Translates the **NumPy** datatype into a **GDAL** datatype integer.

    Parameters
    ----------
    numpy_datatype : np.dtype
        The **NumPy** datatype.

    Returns
    -------
    int
        The GDAL datatype integer.
    """
    assert isinstance(numpy_datatype, (np.dtype, str, int)), f"numpy_datatype must be a numpy.dtype, str or int. Received: {numpy_datatype}"

    parsed = _parse_dtype(numpy_datatype)
    dtype = gdal_array.NumericTypeCodeToGDALTypeCode(parsed)

    return dtype


def _get_default_nodata_value(dtype: Union[np.dtype, str, int]) -> Union[float, int]:
    """
    Returns the default fill value for masked numpy arrays.

    Parameters
    ----------
    dtype : Union[np.dtype, str, int]
        The data type of the array, can be either a numpy dtype object, a string representing a
            data type (e.g. 'float32') or an integer representing a numpy data type (e.g. 5 for 'float32').

    Returns
    -------
    Union[float, int]
        The default fill value for masked numpy arrays.
    """
    assert isinstance(dtype, (np.dtype, str, int)), "numpy_dtype must be a numpy.dtype or string."

    datatypes = {
        "int8": -127,
        "int16": -32767,
        "int32": -2147483647,
        "int64": -9223372036854775807,
        "uint8": 255,
        "uint16": 65535,
        "uint32": 4294967295,
        "uint64": 18446744073709551615,
        "float16": -9999.0,
        "float32": -9999.0,
        "float64": -9999.0,
        "cfloat32": -9999.0,
        "cfloat64": -9999.0,
    }

    test_type = dtype
    if isinstance(dtype, np.dtype):
        test_type = test_type.name
    elif isinstance(dtype, int):
        test_type = _translate_dtype_gdal_to_numpy(dtype).name

    test_type = test_type.lower()

    if test_type in datatypes:
        return datatypes[test_type]

    raise ValueError("Unknown numpy datatype: " + test_type)


def _get_range_for_numpy_datatype(numpy_dtype: Union[str, np.dtype, int]) -> Tuple[Union[int, float], Union[int, float]]:
    """
    Returns the range of values that can be represented by a given numpy dtype.
    `(min_value, max_value)`

    Parameters
    ----------
    numpy_dtype : Union[str, np.dtype]
        The numpy dtype.

    Returns
    -------
    Tuple[Union[int, float], Union[int, float]]
        The range of values that can be represented by a given numpy dtype.
        `(min_value, max_value)`
    """
    assert isinstance(numpy_dtype, (str, np.dtype, int)), "numpy_dtype must be a numpy.dtype or string."

    datatypes = {
        "int8": (-128, 127),
        "int16": (-32768, 32767),
        "int32": (-2147483648, 2147483647),
        "int64": (-9223372036854775808, 9223372036854775807),
        "uint8": (0, 255),
        "uint16": (0, 65535),
        "uint32": (0, 4294967295),
        "uint64": (0, 18446744073709551615),
        "float16": (-6.1e+4, 6.1e+4),
        "bfloat16": (-3.4e+38, 3.4e+38),
        "float32": (-3.4e+38, 3.4e+38),
        "float64": (-1.8e+308, 1.8e+308),
        "cfloat32": (-3.4e+38, 3.4e+38),
        "cfloat64": (-1.8e+308, 1.8e+308),
    }

    test_type = numpy_dtype
    if isinstance(numpy_dtype, np.dtype):
        test_type = test_type.name
    elif isinstance(numpy_dtype, int):
        test_type = _translate_dtype_gdal_to_numpy(numpy_dtype).name

    if test_type in datatypes:
        return datatypes[test_type]

    raise ValueError("Unknown numpy datatype: " + test_type)


def _check_is_value_within_dtype_range(
    value: Union[int, float],
    numpy_dtype: Union[str, np.dtype],
) -> bool:
    """
    Checks if a value is within the range of a numpy datatype.

    Parameters
    ----------
    value : Union[int, float]
        The value to check.

    numpy_dtype : Union[str, np.dtype]
        The numpy dtype.

    Returns
    -------
    bool
        True if the value is within the range of the numpy dtype, otherwise False.
    """
    if value is np.nan:
        return True

    min_val, max_val = _get_range_for_numpy_datatype(numpy_dtype)

    return min_val <= value <= max_val


def _check_is_gdal_dtype_float(gdal_dtype: int) -> bool:
    """
    Checks if a GDAL datatype integer is a floating-point datatype:
    (Float32, Float64, cFloat32, cFloat64)

    Parameters
    ----------
    gdal_dtype : int
        The GDAL datatype integer.

    Returns
    -------
    bool
        True if the GDAL datatype integer is a floating-point datatype, otherwise False.
    """
    assert isinstance(gdal_dtype, int), f"gdal_dtype must be an integer. Received: {gdal_dtype}"

    floats = [gdal.GDT_Float32, gdal.GDT_Float64, gdal.GDT_CFloat32, gdal.GDT_CFloat64]

    if gdal_dtype in floats:
        return True

    return False


def _parse_dtype(
    dtype: Union[str, np.dtype, int, Type[np.int64]]
) -> np.dtype:
    """
    Parses a numpy dtype from a string, numpy dtype, or GDAL datatype integer.

    Parameters
    ----------
    dtype : Union[str, np.dtype, int, type(np.int64)]
        The numpy dtype.

    Returns
    -------
    np.dtype
        The numpy dtype.
    """
    try:
        if isinstance(dtype, str):
            dtype = np.dtype(dtype.lower())
        elif isinstance(dtype, int):
            dtype = _translate_dtype_gdal_to_numpy(dtype)
        elif isinstance(dtype, type(np.int64)):
            dtype = np.dtype(dtype)
        elif isinstance(dtype, np.dtype):
            pass
        else:
            raise ValueError(f"Could not parse dtype: {dtype}")

    except Exception as e:
        raise ValueError(f"Could not parse dtype: {dtype}", e) from None

    return dtype


def _check_is_int_numpy_dtype(
    dtype: Union[str, np.dtype, int, Type[np.int64]],
):
    """
    Checks if a numpy dtype is an integer.

    Parameters
    ----------
    dtype : Union[str, np.dtype, int, type(np.int64)]
        The numpy dtype.

    Returns
    -------
    bool
        True if the numpy dtype is an integer, otherwise False.
    """
    dtype = _parse_dtype(dtype)

    if dtype.kind == "i":
        return True

    return False


def _safe_numpy_casting(
    arr: np.ndarray,
    target_dtype: Union[str, np.dtype, Type[np.int8]],
):
    """
    Safe casting of numpy arrays.
    Clips to min/max of destinations and rounds properly.

    Parameters
    ----------
    arr : np.ndarray
        The array to cast.

    target_dtype : Union[str, np.dtype, type(np.int8)]
        The target dtype.
    
    Returns
    -------
    np.ndarray
        The casted array.
    """
    assert isinstance(arr, np.ndarray), "arr must be a numpy array."
    assert isinstance(target_dtype, (str, np.dtype, type(np.int8))), "target_dtype must be a string or numpy dtype."

    target_dtype = _parse_dtype(target_dtype)

    if arr.dtype == target_dtype:
        return arr

    min_val, max_val = _get_range_for_numpy_datatype(target_dtype.name)
    outarr = np.zeros(arr.shape, dtype=target_dtype)

    if _check_is_int_numpy_dtype(target_dtype) and not _check_is_int_numpy_dtype(arr.dtype):
        outarr[:] = np.rint(arr)
        np.clip(outarr, min_val, max_val, out=outarr)
    else:
        np.clip(arr, min_val, max_val, out=outarr)

    return outarr



""" Create patches from rasters, used for machine learnign applications. """

import sys; sys.path.append("../../")
from typing import Union, List, Tuple, Optional, Callable

import numpy as np
from numba import prange, jit
from buteo.array.utils_array import channel_first_to_last, channel_last_to_first


def _get_kernel_weights(
    tile_size: int = 64,
    edge_distance: int = 5,
    epsilon: float = 1e-7,
) -> np.ndarray:
    """
    Weight a kernel according to how close to an edge a given pixel is.

    Parameters
    ----------
    tile_size : int, optional
        The size of the square kernel. Default: 64

    edge_distance : int, optional
        The distance from the edge to consider for weighting. Default: 5

    epsilon : float, optional
        A small value to prevent division by zero. Default: 1e-7

    Returns
    -------
    np.ndarray
        A 2D NumPy array of shape (tile_size, tile_size) with the kernel weights.
    """
    assert tile_size > 0, "Tile size must be greater than zero."
    assert edge_distance < tile_size // 2, "Edge distance must be less than half the tile size."
    assert edge_distance >= 0, "Edge distance must be greater than or equal to zero."

    arr = np.zeros((tile_size, tile_size), dtype="float32")
    max_dist = edge_distance * 2

    # Iterate through the kernel array
    for idx_y in range(0, arr.shape[0]):
        for idx_x in range(0, arr.shape[1]):

            # Calculate vertical distance to the closest edge
            val_y_top = max(edge_distance - idx_y, 0.0)
            val_y_bot = max((1 + edge_distance) - (tile_size - idx_y), 0.0)
            val_y = val_y_top + val_y_bot

            # Calculate horizontal distance to the closest edge
            val_x_lef = max(edge_distance - idx_x, 0.0)
            val_x_rig = max((1 + edge_distance) - (tile_size - idx_x), 0.0)
            val_x = val_x_lef + val_x_rig

            # Calculate the weight based on the distance to the closest edge
            val = (max_dist - abs(val_y + val_x)) / max_dist

            # Set a minimum weight to avoid division by zero
            if val <= 0.0:
                val = epsilon

            # Assign the calculated weight to the kernel array
            arr[idx_y, idx_x] = val

    return arr


@jit(nopython=True, parallel=True, nogil=True, cache=True)
def _merge_weighted_median(
    arr: np.ndarray,
    arr_weight: np.ndarray,
) -> np.ndarray:
    """
    Calculate the weighted median of a multi-dimensional array along the first axis.
    This is the order (number_of_overlaps, tile_size, tile_size, number_of_bands)
    
    Parameters
    ----------
    arr : np.ndarray
        The input array.
    
    arr_weight : np.ndarray
        The weight array with the same shape as the input array.

    Returns
    -------
    np.ndarray
        A 3D NumPy array of shape (arr.shape[1], arr.shape[2], arr.shape[3]) with the weighted medians.
    """
    ret_arr = np.empty((arr.shape[1], arr.shape[2], arr.shape[3]), dtype="float32")
    ret_arr[:] = np.nan

    # Iterate through the input array
    for idx_y in prange(arr.shape[1]):
        for idx_x in range(arr.shape[2]):
            for idx_band in range(arr.shape[3]):

               # Flatten the input and weight arrays
                values = arr[:, idx_y, idx_x, idx_band].flatten()
                weights = arr_weight[:, idx_y, idx_x, 0].flatten()

                nan_mask = np.where(~np.isnan(values))[0]

                if len(nan_mask) == 0:
                    continue

                values = values[nan_mask]
                weights = weights[nan_mask]

                # Sort the values and weights based on the values
                sort_mask = np.argsort(values)
                sorted_data = values[sort_mask]
                sorted_weights = weights[sort_mask]

                # Calculate the cumulative sum of the sorted weights
                cumsum = np.cumsum(sorted_weights)

                # Normalize the cumulative sum to the range [0, 1]
                intersect = (cumsum - 0.5 * sorted_weights) / cumsum[-1]

                # Interpolate the weighted median and store it in the result array
                ret_arr[idx_y, idx_x, idx_band] = np.interp(0.5, intersect, sorted_data)

    return ret_arr


@jit(nopython=True, parallel=True, nogil=True, cache=True)
def _merge_weighted_average(
    arr: np.ndarray,
    arr_weight: np.ndarray,
) -> np.ndarray:
    """
    Calculate the weighted average of a multi-dimensional array along the last axis.
    
    Parameters
    ----------
    arr : np.ndarray
        The input array.

    arr_weight : np.ndarray
        The weight array with the same shape as the input array.

    Returns
    -------
    np.ndarray
        A 3D NumPy array of shape (arr.shape[0], arr.shape[1], 1) with the weighted averages.
    """
    ret_arr = np.empty((arr.shape[1], arr.shape[2], arr.shape[3]), dtype="float32")
    ret_arr[:] = np.nan

    # Iterate through the input array
    for idx_y in prange(arr.shape[1]):
        for idx_x in range(arr.shape[2]):
            for idx_band in range(arr.shape[3]):

                # Flatten the input and weight arrays
                values = arr[:, idx_y, idx_x, idx_band].flatten()
                weights = arr_weight[:, idx_y, idx_x, 0].flatten()

                nan_mask = np.where(~np.isnan(values))[0]

                if len(nan_mask) == 0:
                    continue

                values = values[nan_mask]
                weights = weights[nan_mask]

                # Calculate the weighted sum and total weight
                weighted_sum = np.nansum(values * weights)
                total_weight = np.nansum(weights)

                # Calculate the weighted average and store it in the result array
                ret_arr[idx_y, idx_x, idx_band] = weighted_sum / total_weight

    return ret_arr

@jit(nopython=True, parallel=True, nogil=True, cache=True)
def _merge_weighted_minmax(
    arr: np.ndarray,
    arr_weight: np.ndarray,
    method="max",
) -> np.ndarray:
    """
    Calculate the weighted min or max of a multi-dimensional array along the last axis.
    
    Parameters
    ----------
    arr : np.ndarray
        The input array.
    
    arr_weight : np.ndarray
        The weight array with the same shape as the input array.
    
    method : str, optional
        The method to use. Either "min" or "max". Default: "max"

    Returns
    -------
    np.ndarray
        A 3D NumPy array of shape (arr.shape[0], arr.shape[1], 1) with the weighted min or max.
    """
    ret_arr = np.empty((arr.shape[1], arr.shape[2], arr.shape[3]), dtype="float32")
    ret_arr[:] = np.nan

    minmax = 0
    if method == "min":
        minmax = 1

    # Iterate through the input array
    for idx_y in prange(arr.shape[1]):
        for idx_x in range(arr.shape[2]):
            for idx_band in range(arr.shape[3]):

                # Flatten the input and weight arrays
                values = arr[:, idx_y, idx_x, idx_band].flatten()
                weights = arr_weight[:, idx_y, idx_x, 0].flatten()

                nan_mask = np.where(~np.isnan(values))[0]

                if len(nan_mask) == 0:
                    continue

                values = values[nan_mask]
                weights = weights[nan_mask]

                weighted = values * weights

                if minmax == 0: # max
                    index = np.nanargmax(weighted)
                else:
                    index = np.nanargmin(weighted)

                value = values[index]

                # Calculate the weighted average and store it in the result array
                ret_arr[idx_y, idx_x, idx_band] = value

    return ret_arr


@jit(nopython=True, parallel=True, nogil=True, cache=True)
def _merge_weighted_olympic(
    arr: np.ndarray,
    arr_weight: np.ndarray,
    level: int = 1,
) -> np.ndarray:
    """
    Calculate the olympic value of a multi-dimensional array along the last axis.
    Using olympic sort, the highest and lowest values are removed from the calculation.
    If level is 1, then the highest and loweest values are removed. If the level is 2,
    then the 2 highest and lowest values are removed, and so on.
    
    Parameters
    ----------
    arr : np.ndarray
        The input array.

    arr_weight : np.ndarray
        The weight array with the same shape as the input array.
    
    level : int, optional
        The level of olympic sort to use. Default: 1

    Returns
    -------
    np.ndarray
        A 3D NumPy array of shape (arr.shape[0], arr.shape[1], 1) with the olympic value.
    """
    ret_arr = np.empty((arr.shape[1], arr.shape[2], arr.shape[3]), dtype="float32")
    ret_arr[:] = np.nan

    required = int((level * 2) + 1)

    # Iterate through the input array
    for idx_y in prange(arr.shape[1]):
        for idx_x in range(arr.shape[2]):
            for idx_band in range(arr.shape[3]):

                # Flatten the input and weight arrays
                values = arr[:, idx_y, idx_x, idx_band].flatten()
                weights = arr_weight[:, idx_y, idx_x, 0].flatten()

                nan_mask = np.where(~np.isnan(values))[0]

                if len(nan_mask) == 0:
                    continue

                values = values[nan_mask]
                weights = weights[nan_mask]

                if len(values) < required: # Take the average of all
                    value = np.mean(values)
                elif len(values) == required: # Take the middle value
                    value = np.sort(values)[level]
                else:
                    sort_olympic = np.argsort(values)[level:-level]
                    sort_weights = weights[sort_olympic] / np.sum(weights[sort_olympic])
                    sort_values = values[sort_olympic]

                    value = np.sum(sort_values * sort_weights)

                # Calculate the weighted average and store it in the result array
                ret_arr[idx_y, idx_x, idx_band] = value

    return ret_arr


@jit(nopython=True, parallel=True, nogil=True)
def _merge_weighted_mad(
    arr: np.ndarray,
    arr_weight: np.ndarray,
    mad_dist: float = 2.0,
) -> np.ndarray:
    """
    Merge an array of predictions using the MAD-merge methodology.
    
    Parameters
    ----------
    arr : np.ndarray
        The input array.

    arr_weight : np.ndarray
        The weight array with the same shape as the input array.

    mad_dist : float, optional
        The MAD distance. Default: 2.0

    Returns
    -------
    np.ndarray
        A 3D NumPy array of shape (arr.shape[0], arr.shape[1], 1) with the MAD-merged values.
    """
    ret_arr = np.empty((arr.shape[1], arr.shape[2], arr.shape[3]), dtype="float32")
    ret_arr[:] = np.nan

    # Iterate through the input array
    for idx_y in prange(arr.shape[1]):
        for idx_x in range(arr.shape[2]):
            for idx_band in range(arr.shape[3]):

                # Flatten the input and weight arrays
                values = arr[:, idx_y, idx_x, idx_band].flatten()
                weights = arr_weight[:, idx_y, idx_x, 0].flatten()

                nan_mask = np.where(~np.isnan(values))[0]

                if len(nan_mask) == 0:
                    continue

                values = values[nan_mask]
                weights = weights[nan_mask]

                # Sort the values and weights based on the values
                sort_mask = np.argsort(values)
                sorted_data = values[sort_mask]
                sorted_weights = weights[sort_mask]

                # Calculate the cumulative sum of the sorted weights and normalize to the range [0, 1]
                cumsum = np.cumsum(sorted_weights)
                intersect = (cumsum - 0.5 * sorted_weights) / cumsum[-1]

                # Interpolate the median
                median = np.interp(0.5, intersect, sorted_data)

                # Calculate the median absolute deviation (MAD)
                mad = np.median(np.abs(median - values))

                # If MAD is zero, store the median in the result array and continue
                if mad == 0.0:
                    ret_arr[idx_y, idx_x, 0] = median
                    continue

                # Calculate the new weights based on the MAD
                new_weights = np.zeros_like(sorted_weights)
                for idx_z in range(sorted_data.shape[0]):
                    new_weights[idx_z] = 1.0 - (np.minimum(np.abs(sorted_data[idx_z] - median) / (mad * mad_dist), 1))

                if np.sum(new_weights) == 0.0:
                    ret_arr[idx_y, idx_x, 0] = median
                    continue

                # Calculate the cumulative sum of the new weights and normalize to the range [0, 1]
                cumsum = np.cumsum(new_weights)
                intersect = (cumsum - 0.5 * new_weights) / cumsum[-1]

                # Interpolate the MAD-merged value and store it in the result array
                ret_arr[idx_y, idx_x, idx_band] = np.interp(0.5, intersect, sorted_data)

    return ret_arr


@jit(nopython=True, nogil=True)
def _unique_values(arr: np.ndarray) -> np.ndarray:
    """
    Find the unique values in a 1D NumPy array.
    
    Parameters
    ----------
    arr : np.ndarray
        The input array.

    Returns
    -------
    np.ndarray
        A 1D NumPy array with the unique values.
    """
    unique = np.empty(arr.size, dtype=arr.dtype)
    unique_count = 0
    for i in range(arr.shape[0]):
        if arr[i] not in unique[:unique_count]:
            unique[unique_count] = arr[i]
            unique_count += 1

    return unique[:unique_count]


@jit(nopython=True, parallel=True, nogil=True)
def _merge_weighted_mode(
    arr: np.ndarray,
    arr_weight: np.ndarray,
) -> np.ndarray:
    """
    Calculate the weighted mode of a multi-dimensional array along the last axis.
    
    Parameters
    ----------
    arr : np.ndarray
        The input array.

    arr_weight : np.ndarray
        The weight array with the same shape as the input array.

    Returns
    -------
    np.ndarray
        A 3D NumPy array of shape (arr.shape[0], arr.shape[1], 1) with the weighted modes.
    """
    ret_arr = np.empty((arr.shape[1], arr.shape[2], arr.shape[3]), dtype=arr.dtype)
    ret_arr[:] = np.nan

    # Iterate through the input array
    for idx_y in prange(arr.shape[1]):
        for idx_x in range(arr.shape[2]):
            for idx_band in range(arr.shape[3]):

                # Flatten the input and weight arrays
                values = arr[:, idx_y, idx_x, idx_band].flatten()
                weights = arr_weight[:, idx_y, idx_x, 0].flatten()

                nan_mask = np.where(~np.isnan(values))[0]

                if len(nan_mask) == 0:
                    continue

                values = values[nan_mask]
                weights = weights[nan_mask]

                # Get unique values and their weighted counts
                unique_vals = _unique_values(values)
                weighted_counts = np.zeros(unique_vals.shape[0])

                # Calculate the weighted sum for each unique value
                for i in range(unique_vals.shape[0]):
                    idxs = np.where(values == unique_vals[i])
                    weighted_counts[i] = np.sum(weights[idxs])

                # Get the index of the maximum weighted sum
                mode_idx = np.argmax(weighted_counts)

                # Store the weighted mode in the result array
                ret_arr[idx_y, idx_x, idx_band] = unique_vals[mode_idx]

    return ret_arr


def _get_offsets(
    tile_size: int,
    n_offsets: int,
):
    """
    Generate a list of offset pairs for a given tile size and number of offsets in y and x dimensions.

    Parameters
    ----------
    tile_size : int
        The size of each tile.

    n_offsets : int
        The desired number of offsets to be calculated in the y and x dimensions.

    Returns
    -------
    List[Tuple[int, int]]
        A list of tuples containing offset pairs for y and x dimensions.
        order is (y, x)
    """
    offset_props = np.arange(0, 1, 1 / (n_offsets + 1))[1:].tolist()
    offsets = [(0, 0)]

    assert tile_size > n_offsets, f"Too many offsets ({n_offsets}) requested for tile_size {tile_size}"

    for val in offset_props:
        offset = int(round((val * tile_size), 2))
        offsets.append((offset, offset))

    return offsets


def _borders_are_necessary(
    arr: np.ndarray,
    tile_size: int,
    offset: List[int],
) -> Tuple[bool, bool]:
    """
    Checks if borders are necessary for the given array.
    Width and height are returned as a tuple.
    order is (y, x).

    Parameters
    ----------
    arr : np.ndarray
        The array to be checked.

    tile_size : int
        The size of each tile.
    
    offset : List[int]
        The offset to be used.
    
    Returns
    -------
    Tuple[bool, bool]
        A tuple containing of borders are needed in (height, width) dims.
    """
    if arr.ndim == 2:
        height, width = arr.shape
    else:
        height, width, _ = arr.shape

    if (height - offset[0]) % tile_size == 0:
        height_border = False
    else:
        height_border = True

    if (width - offset[1]) % tile_size == 0:
        width_border = False
    else:
        width_border = True

    return height_border, width_border


def _borders_are_necessary_list(
    arr: np.ndarray,
    tile_size: int,
    offsets: List[List[int]],
) -> Tuple[bool, bool]:
    """
    Checks if borders are necessary for the given array.
    Width and height are returned as a tuple.

    Parameters
    ----------
    arr : np.ndarray
        The array to be checked.
    
    tile_size : int
        The size of each tile.

    offsets : List[List[int]]
        The offsets to be used.

    Returns
    -------
    Tuple[bool, bool]
        A tuple containing of borders are needed in (height, width) dims.
    """
    height_border = True
    width_border = True

    for offset in offsets:
        offset_height_border, offset_width_border = _borders_are_necessary(
            arr, tile_size, offset
        )
        if not offset_height_border:
            height_border = False

        if not offset_width_border:
            width_border = False

        if not height_border and not width_border:
            break

    return height_border, width_border


def _array_to_patches_single(
    arr: np.ndarray,
    tile_size: int,
    offset: Optional[Union[List[int], Tuple[int, int]]] = None,
) -> np.ndarray:
    """
    Generate patches from an array. Offsets in (y, x) order.

    Parameters
    ----------
    arr : np.ndarray
        The array to be divided into patches.

    tile_size : int
        The size of each tile/patch, e.g., 64 for 64x64 tiles.
    
    offset : Optional[Union[List[int], Tuple[int, int]]], optional
        The y and x offset values for the input array. If not provided, defaults to [0, 0].
    
    Returns
    -------
    np.ndarray
        A numpy array containing the patches.
    """
    assert arr.ndim in [2, 3], "Array must be 2D or 3D"
    assert tile_size > 0, "Tile size must be greater than 0"
    assert offset is None or len(offset) == 2, "Offset must be a list or tuple of length 2"

    # Set default offset to [0, 0] if not provided
    if offset is None:
        offset = [0, 0]

    # Calculate the number of patches in the y and x dimensions
    patches_y = (arr.shape[0] - offset[0]) // tile_size
    patches_x = (arr.shape[1] - offset[1]) // tile_size

    # Calculate cut dimensions for the y and x dimensions
    cut_y = -((arr.shape[0] - offset[0]) % tile_size)
    cut_x = -((arr.shape[1] - offset[1]) % tile_size)

    # Set cut dimensions to None if they are 0
    cut_y = None if cut_y == 0 else cut_y
    cut_x = None if cut_x == 0 else cut_x

    # Reshape the array to separate the patches
    reshaped = arr[offset[0]:cut_y, offset[1]:cut_x].reshape(
        patches_y,
        tile_size,
        patches_x,
        tile_size,
        arr.shape[2],
    )

    # Swap axes to rearrange patches in the correct order
    swaped = reshaped.swapaxes(1, 2)

    # Combine the patches into a single array
    blocks = swaped.reshape(-1, tile_size, tile_size, arr.shape[2])

    return blocks


def _patches_to_array_single(
    patches: np.ndarray,
    shape: Union[List, Tuple],
    tile_size: int,
    offset: Optional[Union[List, Tuple]] = None,
    background_value: Optional[Union[int, float]] = None,
) -> np.ndarray:
    """
    Reconstitute an array from patches.

    Given an array of patches, this function stitches them back together
    to form the original array of the specified shape.

    Parameters
    ----------
    patches : np.ndarray
        A numpy array containing the patches to be stitched together.

    shape : Union[List, Tuple]
        The desired shape of the output array.

    tile_size : int
        The size of each tile/patch, e.g., 64 for 64x64 tiles.

    offset : Optional[Union[List, Tuple]], optional
        The y and x offset values for the target array. If not provided, defaults to [0, 0].

    background_value : Optional[Union[int, float]], optional
        The value to use for the background. If not provided, defaults to np.nan.

    Returns
    -------
    np.ndarray
        A numpy array with the original shape, formed by stitching together the provided patches.
    """
    assert len(shape) in [2, 3], "Shape must be a tuple or list of length 2 or 3"
    assert len(patches.shape) == 4, "Patches must be a 4D array"
    assert patches.shape[1] == tile_size, "Patches must be of size tile_size"
    assert patches.shape[2] == tile_size, "Patches must be of size tile_size"
    assert offset is None or len(offset) == 2, "Offset must be a tuple or list of length 2"

    # Set default offset to [0, 0] if not provided
    if offset is None:
        offset = [0, 0]

    # Create an empty target array of the specified shape
    if background_value is None:
        target = np.full(shape, np.nan, dtype=patches.dtype)
    else:
        target = np.full(shape, background_value, dtype=patches.dtype)

    # Calculate target dimensions
    target_y = ((shape[0] - offset[0]) // tile_size) * tile_size
    target_x = ((shape[1] - offset[1]) // tile_size) * tile_size

    # Calculate cut dimensions
    cut_y = -((shape[0] - offset[0]) % tile_size)
    cut_x = -((shape[1] - offset[1]) % tile_size)

    # Set cut dimensions to None if they are 0
    cut_y = None if cut_y == 0 else cut_y
    cut_x = None if cut_x == 0 else cut_x

    # Calculate the number of tiles in the y and x dimensions
    num_tiles_y = target_y // tile_size
    num_tiles_x = target_x // tile_size

    # Reshape the patches for stitching
    reshape = patches.reshape(
        num_tiles_y,
        num_tiles_x,
        tile_size,
        tile_size,
        patches.shape[3],
        1,
    )

    # Swap axes to rearrange patches in the correct order for stitching
    swap = reshape.swapaxes(1, 2)

    # Combine the patches into a single array
    destination = swap.reshape(
        num_tiles_y * tile_size,
        num_tiles_x * tile_size,
        patches.shape[3],
    )

    # Assign the combined patches to the target array
    target[offset[0]:cut_y, offset[1]:cut_x] = destination

    return target


def _patches_to_weights(
    patches: np.ndarray,
    edge_distance: int,
) -> np.ndarray:
    """ Calculate the weights for each patch based on the distance to the edge. """
    assert len(patches.shape) == 4, "Patches must be a 4D array"
    assert patches.shape[1] == patches.shape[2], "Patches must be square"

    # Calculate the distance to the edge for each patch
    weights = _get_kernel_weights(patches.shape[1], edge_distance)

    # Expand the weights to match the number of patches
    weights = np.repeat(weights[np.newaxis, ...], patches.shape[0], axis=0)[..., np.newaxis]

    return weights


def array_to_patches(
    arr: np.ndarray,
    tile_size: int, *,
    n_offsets: int = 0,
    border_check: bool = True,
    channel_last: bool = True,
) -> np.ndarray:
    """
    Generate patches from an array based on the specified parameters.

    Parameters
    ----------
    arr : np.ndarray
        A numpy array to be divided into patches.

    tile_size : int
        The size of each tile/patch, e.g., 64 for 64x64 tiles.

    n_offsets : int, optional
        The desired number of offsets to be calculated. Default: 0

    border_check : bool, optional
        Whether or not to include border patches. Default: True

    channel_last : bool, optional
        Whether or not the channel dimension is the last dimension. Default: True

    Returns
    -------
    np.ndarray
        The concatenate patches along axis 0. In the order (patches, y, x, channels)
    """
    assert len(arr.shape) == 3, "Array must be 3D"
    if channel_last:
        assert arr.shape[0] >= tile_size, "Array must be larger or equal to tile_size"
        assert arr.shape[1] >= tile_size, "Array must be larger or equal to tile_size"
    else:
        assert arr.shape[1] >= tile_size, "Array must be larger or equal to tile_size"
        assert arr.shape[2] >= tile_size, "Array must be larger or equal to tile_size"

    assert tile_size > 0, "Tile size must be greater than 0"
    assert n_offsets >= 0, "Number of offsets must be greater than or equal to 0"
    assert isinstance(border_check, bool), "Border check must be a boolean"
    assert isinstance(n_offsets, int), "Number of offsets must be an integer"

    if not channel_last:
        arr = channel_first_to_last(arr)

    # Get the list of offsets for both x and y dimensions
    offsets = _get_offsets(tile_size, n_offsets)

    if border_check:
        borders_y, borders_x = _borders_are_necessary_list(arr, tile_size, offsets)

        # TODO: Investigate how to handle smarter. Currently we might get duplicates.
        if borders_y or borders_x:
            offsets.append((0, arr.shape[1] - tile_size))
            offsets.append((arr.shape[0] - tile_size, 0))
            offsets.append((arr.shape[0] - tile_size, arr.shape[1] - tile_size))

    # Initialize an empty list to store the generated patches
    patches = []

    # Iterate through the offsets and generate patches for each offset
    for offset in offsets:
        patches.append(
            _array_to_patches_single(arr, tile_size, offset),
        )

    patches = np.concatenate(patches, axis=0)

    if not channel_last:
        patches = channel_last_to_first(patches)

    return patches


# TODO: Add support for multiple input values in the rasters/arrays
def predict_array(
    arr: np.ndarray,
    callback: Callable[[np.ndarray], np.ndarray], *,
    tile_size: int = 64,
    n_offsets: int = 1,
    border_check: bool = True,
    merge_method: str = "median",
    edge_weighted: bool = True,
    edge_distance: int = 3,
    batch_size: int = 1,
    channel_last: bool = True,
) -> np.ndarray:
    """
    Generate patches from an array. Also outputs the offsets and the shapes of the offsets. Only
    suppors the prediction of single values in the rasters/arrays. 

    Parameters
    ----------
    arr : np.ndarray
        A numpy array to be divided into patches.

    callback : Callable[[np.ndarray], np.ndarray]
        The callback function to be used for prediction. The callback function
        must take a numpy array as input and return a numpy array as output.
    
    tile_size : int, optional
        The size of each tile/patch, e.g., 64 for 64x64 tiles. Default: 64

    n_offsets : int, optional
        The desired number of offsets to be calculated. Default: 1

    border_check : bool, optional
        Whether or not to include border patches. Default: True

    merge_method : str, optional
        The method to use for merging the patches. Valid methods
        are ['mad', 'median', 'mean', 'mode', "min", "max", "olympic1", "olympic2"]. Default: "median"
    
    edge_weighted : bool, optional
        Whether or not to weight the edges patches of patches less than the central parts. Default: True

    edge_distance : int, optional
        The distance from the edge to be weighted less. Usually good to
        adjust this to your maximum convolution kernel size. Default: 3

    batch_size : int, optional
        The batch size to use for prediction. Default: 1

    channel_last : bool, optional
        Whether or not the channel dimension is the last dimension. Default: True

    Returns
    -------
    np.ndarray
        The predicted array.
    """
    assert merge_method in ["mad", "median", "mean", "mode", "min", "max", "olympic1", "olympic2"], "Invalid merge method"
    assert len(arr.shape) == 3, "Array must be 3D"

    if channel_last:
        assert arr.shape[0] >= tile_size, "Array must be larger or equal to tile_size"
        assert arr.shape[1] >= tile_size, "Array must be larger or equal to tile_size"
    else:
        assert arr.shape[1] >= tile_size, "Array must be larger or equal to tile_size"
        assert arr.shape[2] >= tile_size, "Array must be larger or equal to tile_size"

    if not channel_last:
        arr = channel_first_to_last(arr)

    # Get the list of offsets for both x and y dimensions
    offsets = _get_offsets(tile_size, n_offsets)

    if border_check:
        borders_y, borders_x = _borders_are_necessary_list(arr, tile_size, offsets)

        # TODO: Investigate how to handle smarter. Currently we might get duplicates.
        if borders_y or borders_x:
            offsets.append((0, arr.shape[1] - tile_size))
            offsets.append((arr.shape[0] - tile_size, 0))
            offsets.append((arr.shape[0] - tile_size, arr.shape[1] - tile_size))

    # Test output dimensions of prediction
    test_patch = arr[np.newaxis, :tile_size, :tile_size, :]
    test_prediction = callback(test_patch)
    test_shape = test_prediction.shape

    # Initialize an empty list to store the generated patches
    predictions = np.zeros((len(offsets), arr.shape[0], arr.shape[1], test_shape[-1]), dtype=arr.dtype)
    predictions_weights = np.zeros((len(offsets), arr.shape[0], arr.shape[1], 1), dtype=np.float32)

    # Iterate through the offsets and generate patches for each offset
    for idx_i, offset in enumerate(offsets):
        patches = _array_to_patches_single(arr, tile_size, offset)

        offset_predictions = np.empty((patches.shape[0], patches.shape[1], patches.shape[2], test_shape[-1]), dtype=arr.dtype)
        offset_weights = np.empty((patches.shape[0], patches.shape[1], patches.shape[2], 1), dtype=np.float32)
        
        for idx_j in range((patches.shape[0] // batch_size) + (1 if patches.shape[0] % batch_size != 0 else 0)):
            idx_start = idx_j * batch_size
            idx_end = (idx_j + 1) * batch_size

            if not channel_last:
                prediction = channel_first_to_last(
                    callback(
                        channel_last_to_first(patches[idx_start:idx_end, ...])
                    )
                )
            else:
                prediction = callback(patches[idx_start:idx_end, ...])

            if edge_weighted:
                weights = _patches_to_weights(patches[idx_start:idx_end, ...], edge_distance)
            else:
                weights = np.ones((tile_size, tile_size, 1), dtype=np.float32)
                weights = np.repeat(weights[np.newaxis, ...], batch_size, axis=0)

            offset_predictions[idx_start:idx_end, ...] = prediction
            offset_weights[idx_start:idx_end, ...] = weights

        predictions[idx_i, :, :, :] = _patches_to_array_single(offset_predictions, (arr.shape[0], arr.shape[1], test_shape[-1]), tile_size, offset)
        predictions_weights[idx_i, :, :, :] = _patches_to_array_single(offset_weights, (arr.shape[0], arr.shape[1], 1), tile_size, offset, 0.0)

    # Merge the predictions
    if merge_method == "mad":
        predictions = _merge_weighted_mad(predictions, predictions_weights)
    elif merge_method == "median":
        predictions = _merge_weighted_median(predictions, predictions_weights)
    elif merge_method in ["mean", "average", "avg"]:
        predictions = _merge_weighted_average(predictions, predictions_weights)
    elif merge_method == "mode":
        predictions = _merge_weighted_mode(predictions, predictions_weights)
    elif merge_method == "max":
        predictions = _merge_weighted_minmax(predictions, predictions_weights, "max")
    elif merge_method == "min":
        predictions = _merge_weighted_minmax(predictions, predictions_weights, "min")
    elif merge_method == "olympic1":
        predictions = _merge_weighted_olympic(predictions, predictions_weights, 1)
    elif merge_method == "olympic2":
        predictions = _merge_weighted_olympic(predictions, predictions_weights, 2)

    if not channel_last:
        predictions = channel_last_to_first(predictions)

    return predictions


def predict_array_pixel(
    arr: np.ndarray,
    callback: Callable[[np.ndarray], np.ndarray],
) -> np.ndarray:
    """ 
    Predicts an array pixel by pixel.

    Args:
        arr (np.ndarray): A numpy array to be divided into patches.
        callback (function): The callback function to be used for prediction. The callback function
            must take a numpy array as input and return a numpy array as output.
    """
    assert len(arr.shape) == 3, "Array must be 3D"

    reshaped = arr.reshape((arr.shape[0] * arr.shape[1], arr.shape[2]))
    predicted = callback(reshaped)
    predicted = predicted.reshape((arr.shape[0], arr.shape[1], predicted.shape[-1]))

    return predicted



"""
### Resample rasters. ###

Module to resample rasters to a target resolution.
Can uses references from vector or other raster datasets.
"""

# Standard library
import sys; sys.path.append("../../")
from typing import Union, Optional, List

# External
from osgeo import gdal
import numpy as np

# Internal
from buteo.utils import (
    utils_io,
    utils_base,
    utils_gdal,
    utils_path,
    utils_translate,
)
from buteo.raster import core_raster, core_raster_io



def _raster_resample(
    raster: Union[str, gdal.Dataset],
    target_size: Union[List[Union[int, float]], int, float, gdal.Dataset, str],
    out_path: Optional[str] = None,
    *,
    target_in_pixels: bool = False,
    resample_alg: str = "nearest",
    overwrite: bool = True,
    creation_options: Optional[List[str]] = None,
    dtype: Optional[str] = None,
    dst_nodata: Union[float, int, str] = "infer",
    verbose: int = 0,
    ram: float = 0.8,
    ram_max: Optional[int] = None,
    ram_min: Optional[int] = 100,
) -> Union[str, gdal.Dataset]:
    """ Internal. """
    assert isinstance(raster, (gdal.Dataset, str)), f"The input raster must be in the form of a str or a gdal.Dataset: {raster}"

    if out_path is None:
        out_path = utils_path._get_temp_filepath(name="resampled_raster.tif")
    else:
        if not utils_path._check_is_valid_output_filepath(out_path):
            raise ValueError(f"Invalid output path: {out_path}")

    ref = core_raster._raster_open(raster)
    metadata = core_raster._get_basic_metadata_raster(ref)

    if isinstance(target_size, (gdal.Dataset, str)):
        x_res, y_res = utils_gdal._get_raster_size(target_size)
        x_pixels = None
        y_pixels = None
    elif target_in_pixels:
        x_res = None
        y_res = None

        if isinstance(target_size, (list, tuple)):
            assert len(target_size) == 2, f"Invalid target_size. {target_size}"
            x_pixels = target_size[0]
            y_pixels = target_size[1]
        elif isinstance(target_size, (int, float)):
            x_pixels = target_size
            y_pixels = target_size
        else:
            raise RuntimeError(f"Invalid target_size. {target_size}")
    else:
        x_pixels = None
        y_pixels = None

        if isinstance(target_size, (list, tuple)):
            assert len(target_size) == 2, f"Invalid target_size. {target_size}"
            x_res = target_size[0]
            y_res = target_size[1]
        elif isinstance(target_size, (int, float)):
            x_res = target_size
            y_res = target_size
        else:
            raise RuntimeError(f"Invalid target_size. {target_size}")

    out_format = utils_gdal._get_raster_driver_name_from_path(out_path)

    src_nodata = metadata["nodata_value"]
    out_nodata = None
    if dst_nodata == "infer":
        dst_nodata = src_nodata
    else:
        assert isinstance(dst_nodata, (int, float, type(None))), "dst_nodata must be an int, float, 'infer', or 'None'"
        out_nodata = dst_nodata

    if dtype is None:
        dtype = metadata["dtype"]

    if out_nodata is not None and not utils_translate._check_is_value_within_dtype_range(out_nodata, dtype):
        raise ValueError(f"Invalid nodata value for datatype. value: {out_nodata}, dtype: {dtype}")

    utils_path._delete_if_required(out_path, overwrite)

    if verbose == 0:
        gdal.PushErrorHandler("CPLQuietErrorHandler")

    options = gdal.WarpOptions(
        format=out_format,
        width=x_pixels,
        height=y_pixels,
        xRes=x_res,
        yRes=y_res,
        outputType=utils_translate._translate_dtype_numpy_to_gdal(dtype),
        resampleAlg=utils_translate._translate_resample_method(resample_alg),
        creationOptions=utils_gdal._get_default_creation_options(creation_options),
        srcNodata=metadata["nodata_value"],
        dstNodata=out_nodata,
        multithread=True,
        warpMemoryLimit=utils_gdal._get_dynamic_memory_limit(ram, min_mb=ram_min, max_mb=ram_max),
    )

    resampled = gdal.Warp(out_path, ref, options=options)

    if verbose == 0:
        gdal.PopErrorHandler()

    if resampled is None:
        raise RuntimeError(f"Error while resampling raster: {out_path}")

    return out_path


def raster_resample(
    raster,
    target_size,
    out_path=None,
    *,
    target_in_pixels=False,
    resample_alg="nearest",
    creation_options=None,
    dtype=None,
    dst_nodata="infer",
    prefix="",
    suffix="",
    overwrite=True,
    add_uuid=False,
    add_timestamp: bool = False,
    verbose: int = 0,
    ram: float = 0.8,
    ram_max: Optional[int] = None,
    ram_min: Optional[int] = 100,
):
    """
    Resampled raster(s) given a target size.
    Beware, if your input is in latitude and longitude, you'll need to specify the target_size in degrees as well.

    Parameters
    ----------
    raster : str/gdal.Dataset/list
        The input raster(s) to resample.
    
    target_size : str/int/ogr.DataSource/gdal.Dataset
        The desired resolution for the resampled raster(s), in the same unit as the raster projection.
        x_res, y_res - or x_pixels, y_pixels if target_in_pixels is True.

    target_in_pixels : bool, optional
        If True, interprets target_size as the number of pixels, default: False

    out_path : str, optional
        The output path for the resampled raster(s). If not provided, the output path is inferred from the input raster(s), default: None

    resample_alg : str, optional
        The resampling algorithm to use, default: "nearest"

    creation_options : list, optional
        A list of creation options for the output raster(s), default: None

    dtype : str, optional
        The output data type, default: None

    dst_nodata : str/int/float, optional
        The nodata value for the output raster(s), default: "infer"

    prefix : str, optional
        A prefix to add to the output path, default: ""

    suffix : str, optional
        A suffix to add to the output path, default: ""

    add_uuid : bool, optional
        If True, adds a uuid to the output path, default: False

    add_timestamp : bool, optional
        If True, adds a timestamp to the output path, default: False

    verbose : int, optional
        The verbosity level, default: 0

    ram : float, optional
        The amount of RAM to use for the resampling, default: 0.8 (80% of the total RAM)

    ram_max : int, optional
        The maximum amount of RAM to use for the resampling, default: None

    ram_min : int, optional
        The minimum amount of RAM to use for the resampling, default: 100

    overwrite : bool, optional
        If True, overwrites the output raster(s) if it/they already exist, default: True

    Returns
    -------
    str/list[str]
        The output path(s) of the resampled raster(s).
    """
    utils_base._type_check(raster, [str, gdal.Dataset, [str, gdal.Dataset]], "raster")
    utils_base._type_check(target_size, [tuple, [int, float], int, float, str, gdal.Dataset], "target_size")
    utils_base._type_check(target_in_pixels, [bool], "target_in_pixels")
    utils_base._type_check(out_path, [str, [str], None], "out_path")
    utils_base._type_check(resample_alg, [str], "resample_alg")
    utils_base._type_check(overwrite, [bool], "overwrite")
    utils_base._type_check(creation_options, [[str], None], "creation_options")
    utils_base._type_check(dst_nodata, [str, int, float, None], "dst_nodata")
    utils_base._type_check(dtype, [str, None, np.dtype, type(np.int8)], "dtype")
    utils_base._type_check(prefix, [str], "prefix")
    utils_base._type_check(suffix, [str], "postfix")
    utils_base._type_check(add_uuid, [bool], "add_uuid")
    utils_base._type_check(add_timestamp, [bool], "add_timestamp")
    utils_base._type_check(verbose, [int], "verbose")
    utils_base._type_check(ram, [int, float], "ram")
    utils_base._type_check(ram_max, [int, float, None], "ram_max")
    utils_base._type_check(ram_min, [int, float, None], "ram_min")

    input_is_list = isinstance(raster, list)

    input_rasters = utils_io._get_input_paths(raster, "raster")
    output_rasters = utils_io._get_output_paths(
        input_rasters,
        out_path,
        add_uuid=add_uuid,
        add_timestamp=add_timestamp,
        prefix=prefix,
        suffix=suffix,
        overwrite=overwrite,
    )

    utils_path._delete_if_required_list(output_rasters, overwrite)

    resampled_rasters = []
    for idx, in_raster in enumerate(input_rasters):
        resampled_rasters.append(
            _raster_resample(
                in_raster,
                target_size,
                target_in_pixels=target_in_pixels,
                out_path=output_rasters[idx],
                resample_alg=resample_alg,
                overwrite=overwrite,
                creation_options=creation_options,
                dtype=dtype,
                dst_nodata=dst_nodata,
                ram=ram,
                ram_max=ram_max,
                ram_min=ram_min,
            )
        )

    if input_is_list:
        return resampled_rasters

    return resampled_rasters[0]


def resample_array(
    arr,
    target_shape_pixels,
    resample_alg="bilinear",
    channel_last=True,
):
    """ Resample a numpy array using the GDAL algorithms. """
    utils_base._type_check(arr, [np.ndarray, np.ma.MaskedArray], "arr")
    utils_base._type_check(target_shape_pixels, [tuple, [int, float]], "target_shape_pixels")
    utils_base._type_check(resample_alg, [str], "resample_alg")
    utils_base._type_check(channel_last, [bool], "channel_last")

    assert len(arr.shape) in [2, 3], f"Invalid array shape: {arr.shape}"
    assert len(target_shape_pixels) in [2, 3], f"Invalid target_shape_pixels: {target_shape_pixels}"

    if len(target_shape_pixels) == 3:
        if channel_last:
            target_shape_pixels = (target_shape_pixels[0], target_shape_pixels[1])
        else:
            target_shape_pixels = (target_shape_pixels[1], target_shape_pixels[2])
    else:
        raise ValueError(f"Invalid target_shape_pixels: {target_shape_pixels}")

    arr_as_raster = core_raster_io.raster_create_from_array(arr, channel_last=channel_last)
    resampled = _raster_resample(
        arr_as_raster,
        target_shape_pixels,
        target_in_pixels=True,
        resample_alg=resample_alg,
    )
    out_arr = core_raster_io.raster_to_array(resampled)

    utils_gdal.delete_dataset_if_in_memory(arr_as_raster)
    utils_gdal.delete_dataset_if_in_memory(resampled)

    return out_arr



from typing import Union, List, Tuple, Optional

import numpy as np


class MultiArray:
    def __init__(self,
        array_list: List[Union[np.ndarray, np.memmap]],
        shuffle: bool = False,
        random_sampling: bool = False,
        seed: int = 42,
        _idx_start: Optional[int] = None,
        _idx_end: Optional[int] = None,
        _is_subarray: bool = False
    ):
        """
        This is a class that takes in a tuple of list of arrays and glues them together
        without concatenating them. This is useful for when you have a large
        dataset that you want to load into memory, but you don't want to
        concatenate them because that would take up too much memory.

        The function works for saved numpy arrays loading using mmap_mode="r".

        Parameters
        ----------
        array_list : List[Union[np.ndarray, np.memmap]]
            A list of numpy arrays to load as one. Can be mmaped or not.
        
        shuffle : bool, default: False
            Whether to shuffle the data or not. Cannot be used together with random_sampling.
            This is different from random sampling in that it ensures that all data will be used.
            It creates an index array that is shuffled and then uses that to index the MultiArray.
        
        random_sampling : bool, default: False
            Whether to use random sampling or not. If True, the returned data will be randomly sampled from the MultiArray.
            Cannot be used together with shuffling.
            Does not ensure that all data will be used, but it will be randomly (uniform) sampled.

        Returns
        -------
        MultiArray
            A class that can handle multiple memmapped numpy arrays.

        Examples
        --------
        ```python
        >>> from glob import glob
        >>> folder = "./data_patches/"

        >>> patches = sorted(glob(folder + "train*.npy"))
        >>> multi_array = MultiArray([np.load(p, mmap_mode="r") for p in patches])
        >>> single_image = multi_array[0]

        >>> print(single_image.shape)
        (128, 128, 10)
        >>> print(len(multi_array))
        32
        ```
        """
        self.array_list = array_list
        self.is_subarray = _is_subarray
        
        assert isinstance(self.array_list, list), "Input should be a list of numpy arrays."
        assert len(self.array_list) > 0, "Input list is empty. Please provide a list with numpy arrays."
        assert all(isinstance(item, (np.ndarray, np.memmap)) for item in self.array_list), "Input list should only contain numpy arrays."

        self.cumulative_sizes = np.cumsum([0] + [arr.shape[0] for arr in self.array_list])

        self._idx_start = int(_idx_start) if _idx_start is not None else 0
        self._idx_end = int(_idx_end) if _idx_end is not None else int(self.cumulative_sizes[-1])

        assert isinstance(self._idx_start, int), "Minimum length should be an integer."
        assert isinstance(self._idx_end, int), "Maximum length should be an integer."
        assert self._idx_start < self._idx_end, "Minimum length should be smaller than maximum length."

        self.total_length = int(min(self.cumulative_sizes[-1], self._idx_end - self._idx_start))  # Store length for faster access

        if shuffle and random_sampling:
            raise ValueError("Cannot use both shuffling and resevoir sampling at the same time.")

        # Shuffling
        self.seed = seed
        self.shuffle = shuffle
        self.shuffle_indices = None
        self.random_sampling = random_sampling
        self.rng = np.random.default_rng(seed)

        if self.shuffle:
            self.shuffle_indices = self.rng.permutation(range(self._idx_start, self._idx_end))

    def _load_item(self, idx: int):
        """ Load an item from the array list. """
        array_idx = np.searchsorted(self.cumulative_sizes, idx, side='right') - 1

        calculated_idx = idx - self.cumulative_sizes[array_idx]
        if calculated_idx < 0 or calculated_idx >= self.array_list[array_idx].shape[0]:
            raise IndexError(f'Index {idx} out of bounds for MultiArray with length {self.__len__()}')

        output = self.array_list[array_idx][calculated_idx]

        return output


    def set_shuffle_index(self, shuffle_indices: np.ndarray) -> None:
        """ Set the shuffle indices and enable shuffling. """
        self.shuffle_indices = shuffle_indices
        self.shuffle = True
        assert len(self.shuffle_indices) == len(self), "Length of shuffle indices should be equal to the length of the MultiArray."

    
    def get_shuffle_index(self) -> np.ndarray:
        """ Get the shuffle indices. """
        return self.shuffle_indices
    
    
    def shuffle_index(self) -> None:
        """ Shuffle the MultiArray. """
        self.shuffle_indices = self.rng.permutation(range(self._idx_start, self._idx_end))
        self.shuffle = True

    
    def disable_shuffle(self) -> None:
        """ Disable shuffling. """
        self.shuffle = False


    def __getitem__(self, idx: int):
        if idx < 0:
            idx = self._idx_end + idx

        if self.random_sampling:
            idx = np.random.randint(self._idx_start, self._idx_end)
        elif self.shuffle:
            idx = self.shuffle_indices[idx]
        else:
            idx = idx + self._idx_start

        if idx < self._idx_start or idx >= self._idx_end:
            raise IndexError("Index out of range")

        return self._load_item(idx)


    def split(self, split_point: Union[int, float]) -> Tuple["MultiArray", "MultiArray"]:
        """
        Split the MultiArray into two MultiArrays.

        Parameters
        ----------
        split_point : int or float
            The split point for the multi array. If float, it will be interpreted as a percentage.
            If int, it will be interpreted as an index.

        Returns
        -------
        MultiArray, MultiArray
            Two MultiArray objects. (before_split_point, after_split_point)
        """
        if isinstance(split_point, float):
            split_point = int(split_point * len(self))

        if split_point > len(self):
            raise ValueError("Split point is larger than the length of the MultiArray.")
            
        if self.is_subarray:
            raise ValueError("Cannot split an array that has already been split.")

        before_split_point = MultiArray(self.array_list, shuffle=False, random_sampling=self.random_sampling, seed=self.seed, _idx_start=self._idx_start, _idx_end=self._idx_start + split_point, _is_subarray=True)
        after_split_point = MultiArray(self.array_list, shuffle=False, random_sampling=self.random_sampling, seed=self.seed, _idx_start=self._idx_start + split_point, _idx_end=self._idx_end, _is_subarray=True)

        return before_split_point, after_split_point


    def __iter__(self):
        for i in range(self.__len__()):
            yield self[i]


    def __len__(self):
        return self.total_length
